import networkx as nx
import pytest

import skmine.graph.graphmdl.utils as utils
from ..candidate import Candidate
from ..code_table import CodeTable
from ..code_table_row import CodeTableRow
from ..graph_mdl import GraphMDL
from ..label_codes import LabelCodes

pattern = nx.DiGraph()
pattern.add_node(1)
pattern.add_node(2)
pattern.add_edge(1, 2)

row = CodeTableRow(pattern)

""" Code Table Row"""


def test_pattern():
    assert len(row.pattern().nodes()) == 2
    assert len(row.pattern().edges()) == 1


def test_pattern_code():
    assert row.pattern_usage() is None


def test_set_pattern_code():
    row.set_pattern_usage(3.0)
    assert row.pattern_usage() == 3.0


def test_pattern_port_code():
    assert row.pattern_port_usage() is None


def test_embeddings():
    assert len(row.embeddings()) == 0


def test_set_embeddings():
    row.set_embeddings([0, 1, 2, 3])
    assert len(row.embeddings()) == 4
    assert row.embeddings()[0] == 0
    assert row.embeddings()[3] == 3


""" Label codes"""


def init_label_codes_test():
    res = dict()
    g = nx.DiGraph()
    g.add_nodes_from(range(1, 9))
    g.add_edge(2, 1, label='a')
    g.add_edge(4, 1, label='a')
    g.add_edge(6, 1, label='a')
    g.add_edge(6, 8, label='a')
    g.add_edge(8, 6, label='a')
    g.add_edge(1, 3, label='b')
    g.add_edge(1, 5, label='b')
    g.add_edge(1, 7, label='b')
    g.nodes[1]['label'] = 'y'
    g.nodes[2]['label'] = 'x'
    g.nodes[3]['label'] = 'z'
    g.nodes[4]['label'] = 'x'
    g.nodes[5]['label'] = 'z'
    g.nodes[6]['label'] = 'x'
    g.nodes[7]['label'] = 'z'
    g.nodes[8]['label'] = 'w', 'x'

    graph = g
    lc = LabelCodes(graph)
    res['st'] = lc
    res['graph'] = graph
    return res


def test_total_label():
    st = init_label_codes_test()['st']
    assert st.total_label() == 17.0


def test_vertex_st():
    st = init_label_codes_test()['st']
    assert pytest.approx(st.vertex_lc()['y'], rel=1e-01) == 4.09
    assert pytest.approx(st.vertex_lc()['x'], rel=1e-01) == 2.09
    assert pytest.approx(st.vertex_lc()['z'], rel=1e-01) == 2.50
    assert pytest.approx(st.vertex_lc()['w'], rel=1e-01) == 4.09


def test_edges_st():
    st = init_label_codes_test()['st']
    assert pytest.approx(st.edges_lc()['a'], rel=1e-01) == 1.77
    assert pytest.approx(st.edges_lc()['b'], rel=1e-01) == 2.50


def test_encode():
    st = init_label_codes_test()['st']
    g1 = nx.DiGraph()
    g1.add_nodes_from(range(1, 3))
    g1.add_edge(1, 2, label='a')
    g1.nodes[1]['label'] = 'x'
    assert pytest.approx(st.encode(g1), rel=1e-01) == 21.44


def test_encode_singleton_vertex():
    st = init_label_codes_test()['st']
    assert pytest.approx(st.encode_singleton_vertex('x'), rel=1e-01) == 12.67


def test_encode_singleton_edge():
    st = init_label_codes_test()['st']
    assert pytest.approx(st.encode_singleton_edge('a'), rel=1e-01) == 14.35


""" Utils test"""


def init_graph_to_utils_test():
    res = dict()
    g = nx.DiGraph()
    g.add_nodes_from(range(1, 9))
    g.add_edge(2, 1, label='a')
    g.add_edge(4, 1, label='a')
    g.add_edge(6, 1, label='a')
    g.add_edge(6, 8, label='a')
    g.add_edge(8, 6, label='a')
    g.add_edge(1, 3, label='b')
    g.add_edge(1, 5, label='b')
    g.add_edge(1, 7, label='b')
    g.nodes[1]['label'] = 'y'
    g.nodes[2]['label'] = 'x'
    g.nodes[3]['label'] = 'z'
    g.nodes[4]['label'] = 'x'
    g.nodes[5]['label'] = 'z'
    g.nodes[6]['label'] = 'x'
    g.nodes[7]['label'] = 'z'
    g.nodes[8]['label'] = 'w', 'x'
    label_codes = LabelCodes(g)
    res['graph'] = g
    res['st'] = label_codes
    return res


def init_multi_graph():
    graph = nx.MultiDiGraph()
    graph.add_node(1, label="Book")
    graph.add_node(2, label="Book")
    graph.add_node(3, label="Book")
    graph.add_node(4)
    graph.nodes[4]['label'] = 'xsd:string', 'Value:Alice'
    graph.add_node(5, label="Person")
    graph.add_node(6, label="Person")
    graph.add_node(7)
    graph.nodes[7]['label'] = 'xsd:string', 'Value:Bob'
    graph.add_node(8, label="City")
    graph.add_node(9, label="City")
    graph.add_node(10, label="Monument")
    graph.add_node(11, label="Monument")
    graph.add_node(12)
    graph.add_node(13)
    graph.nodes[12]['label'] = 'xsd:integer', 'Value:123'
    graph.nodes[13]['label'] = 'xsd:integer', 'Value:123'
    graph.add_edge(1, 5, label='author')
    graph.add_edge(2, 5, label='author')
    graph.add_edge(3, 5, label='author')
    graph.add_edge(3, 6, label='author')
    graph.add_edge(5, 4, label='name')
    graph.add_edge(5, 8, label='born_in')
    graph.add_edge(5, 8, label='died_in')
    graph.add_edge(6, 7, label='name')
    graph.add_edge(6, 9, label='born_in')
    graph.add_edge(6, 9, label='died_in')
    graph.add_edge(10, 8, label='is_located')
    graph.add_edge(11, 8, label='is_located')
    graph.add_edge(10, 12, label='height')
    graph.add_edge(10, 11, label='near')
    graph.add_edge(11, 10, label='near')
    graph.add_edge(11, 13, label='height')
    label_codes = LabelCodes(graph)
    ct = CodeTable(label_codes, graph)
    ct.cover()

    p1 = nx.MultiDiGraph()
    p1.add_node(1)
    p1.add_node(2, label='Monument')
    p1.add_node(3, label='Monument')
    p1.add_node(4)
    p1.add_node(5)
    p1.nodes[4]['label'] = 'xsd:integer', 'Value:123'
    p1.nodes[5]['label'] = 'xsd:integer', 'Value:123'
    p1.add_edge(2, 1, label='is_located')
    p1.add_edge(3, 1, label='is_located')
    p1.add_edge(2, 4, label='height')
    p1.add_edge(2, 3, label='near')
    p1.add_edge(3, 2, label='near')
    p1.add_edge(3, 5, label='height')
    row1 = CodeTableRow(p1)

    p2 = nx.MultiDiGraph()
    p2.add_node(1)
    p2.add_node(2, label='City')
    p2.add_node(3, label='xsd:string')
    p2.add_edge(1, 3, label='name')
    p2.add_edge(1, 2, label='born_in')
    p2.add_edge(1, 2, label='died_in')
    row2 = CodeTableRow(p2)

    p3 = nx.MultiDiGraph()
    p3.add_node(1, label='Book')
    p3.add_node(2, label='Person')
    p3.add_edge(1, 2, label='author')
    row3 = CodeTableRow(p3)
    return {
        'graph': graph, 'label_codes': label_codes, 'ct': ct,
        'row1': row1, 'row2': row2, 'row3': row3}


def test_prequential_code():
    assert utils.prequential_code([0, 0]) == 0.0
    assert utils.prequential_code([1, 0]) == 1.0
    assert pytest.approx(utils.prequential_code([1, 1, 1, 1]), rel=1e-01) == 10.91
    assert pytest.approx(utils.prequential_code([2, 2, 4]), rel=1e-01) == 15.15
    assert pytest.approx(utils.prequential_code([10, 8, 0, 0, 0, 0, 0]), rel=1e-01) == 29.30
    assert pytest.approx(utils.prequential_code([100, 500, 4, 10, 0]), rel=1e-01) == 515.90


def test_count_edge_label():
    graph = init_graph_to_utils_test()['graph']
    assert len(utils.count_edge_label(graph).items()) == 2
    assert utils.count_edge_label(graph)['a'] == 5
    assert utils.count_edge_label(graph)['b'] == 3


def test_count_vertex_label():
    graph = init_graph_to_utils_test()['graph']
    assert len(utils.count_vertex_label(graph).items()) == 4
    assert utils.count_vertex_label(graph)['x'] == 4
    assert utils.count_vertex_label(graph)['y'] == 1
    assert utils.count_vertex_label(graph)['z'] == 3
    assert utils.count_vertex_label(graph)['w'] == 1


def test_get_total_label():
    graph = init_graph_to_utils_test()['graph']
    assert utils.get_total_label(graph) == 17


def test_binomial():
    with pytest.raises(ValueError):
        utils.binomial(2, 5)

    assert utils.binomial(2, 0) == 1
    assert utils.binomial(4, 3) == 4
    assert utils.binomial(4, 2) == 6


def test_universal_integer_encoding():
    with pytest.raises(ValueError):
        utils.universal_integer_encoding(0)

    assert utils.universal_integer_encoding(1) == 1


def test_universal_integer_encoding_with0():
    with pytest.raises(ValueError):
        utils.universal_integer_encoding_with0(-1)

        assert utils.universal_integer_encoding_with0(1) == 1


def test_utils_encode():
    standard_table = init_graph_to_utils_test()['st']
    graph = init_graph_to_utils_test()['graph']
    g1 = nx.DiGraph()
    g1.add_nodes_from(range(1, 3))
    g1.add_edge(1, 2, label='a')
    g1.nodes[1]['label'] = 'x'

    p5 = nx.DiGraph()
    p5.add_node(1, label='x')
    p5.add_node(2, label='y')
    p5.add_edge(1, 2, label='a')
    # print(utils.encode(p5, standard_table))

    assert pytest.approx(utils.encode(g1, standard_table), rel=1e-01) == 21.44
    assert pytest.approx(utils.encode(graph, standard_table), rel=1e-01) == 111.76


def test_encode_vertex_singleton():
    standard_table = init_graph_to_utils_test()['st']
    with pytest.raises(ValueError):
        utils.encode_vertex_singleton(standard_table, '')

    assert pytest.approx(utils.encode_vertex_singleton(standard_table, 'x'), rel=1e-01) == 12.67
    assert pytest.approx(utils.encode_vertex_singleton(standard_table, 'y'), rel=1e-01) == 14.67
    assert pytest.approx(utils.encode_vertex_singleton(standard_table, 'z'), rel=1e-01) == 13.09
    assert pytest.approx(utils.encode_vertex_singleton(standard_table, 'w'), rel=1e-01) == 14.67


def test_encode_edge_singleton():
    standard_table = init_graph_to_utils_test()['st']
    with pytest.raises(ValueError):
        utils.encode_edge_singleton(standard_table, '')

    assert pytest.approx(utils.encode_edge_singleton(standard_table, 'a'), rel=1e-01) == 14.35
    assert pytest.approx(utils.encode_edge_singleton(standard_table, 'b'), rel=1e-01) == 15.09


def test_encode_singleton():
    standard_table = init_graph_to_utils_test()['st']
    with pytest.raises(ValueError):
        utils.encode_singleton(standard_table, 0, 'a')

    assert pytest.approx(utils.encode_singleton(standard_table, 2, 'a'), rel=1e-01) == 14.35
    assert pytest.approx(utils.encode_singleton(standard_table, 1, 'x'), rel=1e-01) == 12.67

    # Test for graph


def init_ng_to_utils_test():
    res = dict()

    ng = nx.Graph()
    ng.add_nodes_from(range(1, 6))
    ng.add_edge(1, 2, label='e')
    ng.add_edge(2, 3, label='e')
    ng.add_edge(2, 4, label='e')
    ng.add_edge(5, 2, label='e')
    ng.nodes[1]['label'] = 'A'
    ng.nodes[2]['label'] = 'A'
    ng.nodes[3]['label'] = 'B'
    ng.nodes[4]['label'] = 'B'
    ng.nodes[5]['label'] = 'A'
    res['ng'] = ng
    pattern = nx.Graph()
    pattern.add_nodes_from(range(1, 3))
    pattern.add_edge(1, 2, label='e')
    pattern.nodes[1]['label'] = 'A'
    res['pattern'] = pattern
    return res


def test_get_embeddings():
    ng = init_ng_to_utils_test()['ng']
    pattern = init_ng_to_utils_test()['pattern']
    ng2 = nx.DiGraph()
    ng2.add_nodes_from(range(1, 6))
    ng2.add_edge(1, 2, label='e')
    ng2.add_edge(2, 3, label='e')
    ng2.add_edge(2, 4, label='e')
    ng2.add_edge(5, 2, label='e')
    ng2.nodes[1]['label'] = 'A'
    ng2.nodes[2]['label'] = 'A'
    ng2.nodes[3]['label'] = 'B'
    ng2.nodes[4]['label'] = 'B'
    ng2.nodes[5]['label'] = 'A'

    ngp = nx.DiGraph()
    ngp.add_nodes_from(range(1, 3))
    ngp.add_edge(1, 2, label='e')
    ngp.nodes[1]['label'] = 'A'
    # first test for digraph
    assert len(utils.get_embeddings(ngp, ng2)) == 4
    assert utils.get_embeddings(ngp, ng2)[0][1] == 1
    assert utils.get_embeddings(ngp, ng2)[0][2] == 2

    embed = utils.get_embeddings(pattern, ng)
    assert len(utils.get_embeddings(pattern, ng)) == 6
    assert utils.get_embeddings(pattern, ng)[5][5] == 1
    assert utils.get_embeddings(pattern, ng)[5][2] == 2

    test1 = nx.Graph()
    ptest1 = nx.Graph()
    test1.add_node(1)
    test1.nodes[1]['label'] = 'w', 'x', 'y'
    ptest1.add_node(1)
    ptest1.nodes[1]['label'] = 'w', 'x'

    assert len(utils.get_embeddings(ptest1, test1)) != 0
    assert len(utils.get_embeddings(test1, ptest1)) == 0

    test2 = nx.Graph()
    test2.add_node(1)
    ptest2 = nx.Graph()
    ptest2.add_node(1, label='x')
    assert len(utils.get_embeddings(ptest2, test2)) == 0
    assert len(utils.get_embeddings(test2, ptest2)) != 0


def test_is_vertex_singleton():
    g1 = nx.DiGraph()
    g1.add_node(1, label='a')
    assert utils.is_vertex_singleton(g1) is True

    g2 = nx.Graph()
    g2.add_node(1, label='a')
    assert utils.is_vertex_singleton(g2) is True

    g3 = nx.Graph()
    g3.add_nodes_from(range(1, 3))
    assert utils.is_vertex_singleton(g3) is False

    g4 = nx.DiGraph()
    g4.add_node(1)
    g4.nodes[1]['label'] = 'a', 'b'
    assert utils.is_vertex_singleton(g4) is False


def test_is_edge_singleton():
    g1 = nx.DiGraph()
    g1.add_node(1, label='a')
    assert utils.is_edge_singleton(g1) is False

    g2 = nx.Graph()
    g2.add_node(1)
    g2.add_node(2, label='a')
    assert utils.is_edge_singleton(g2) is False

    g4 = nx.Graph()
    g4.add_node(1)
    g4.add_node(2)
    g4.add_edge(1, 2)
    g4[1][2]['label'] = 'a', 'b'
    print(bool(utils.count_edge_label(g4)))
    assert utils.is_edge_singleton(g4) is False

    g5 = nx.DiGraph()
    g5.add_node(1)
    g5.add_node(2)
    g5.add_edge(1, 2)
    g5[1][2]['label'] = 'a'
    assert utils.is_edge_singleton(g5) is True


def test_get_support():
    ng = init_ng_to_utils_test()['ng']
    pattern = init_ng_to_utils_test()['pattern']
    graph = init_graph_to_utils_test()['graph']
    pattern1 = nx.DiGraph()
    pattern1.add_nodes_from(range(1, 4))
    pattern1.nodes[1]['label'] = 'x'
    pattern1.nodes[2]['label'] = 'y'
    pattern1.nodes[3]['label'] = 'z'
    pattern1.add_edge(1, 2, label='a')
    pattern1.add_edge(2, 3, label='b')

    assert utils.get_support(utils.get_embeddings(pattern1, graph)) == 1
    assert utils.get_support(utils.get_embeddings(pattern, ng)) == 2


def test_get_label_index():
    values = ('a', 'b')
    assert utils.get_label_index('a', values) == 0

    with pytest.raises(ValueError):
        utils.get_label_index('c', values)


def test_get_node_label():
    test = nx.Graph()
    test.add_node(1)
    test.nodes[1]['label'] = 'a', 'b'

    with pytest.raises(ValueError):
        utils.get_node_label(1, 6, test)
        utils.get_node_label(2, 0, test)

    assert utils.get_node_label(1, 0, test) == 'a'


def test_get_edge_label():
    test = nx.Graph()
    test.add_nodes_from(range(1, 3))
    test.add_edge(1, 2, label='a')
    assert utils.get_edge_label(1, 2, test) == 'a'

    test.add_edge(2, 3)
    with pytest.raises(ValueError):
        utils.get_edge_label(2, 3, test)
        utils.get_edge_label(2, 4, test)


def test_is_without_edge():
    test = nx.Graph()
    test.add_node(1)
    test.add_node(2)
    assert utils.is_without_edge(test) is True

    test.add_edge(1, 2)
    assert utils.is_without_edge(test) is False


def test_get_edge_in_embedding():
    graph = init_graph_to_utils_test()['graph']
    p7 = nx.DiGraph()
    p7.add_node(1, label='x')
    p7.add_node(2)
    p7.add_node(3)
    p7.add_edge(1, 2, label='a')
    p7.add_edge(1, 3, label='a')

    embeddings = utils.get_embeddings(p7, graph)
    assert list(utils.get_edge_in_embedding(embeddings[0], p7))[0][0] == 1
    assert list(utils.get_edge_in_embedding(embeddings[0], p7))[0][1] == 2
    assert list(utils.get_edge_in_embedding(embeddings[0], p7))[1][0] == 1
    assert list(utils.get_edge_in_embedding(embeddings[0], p7))[1][1] == 3


def test_get_key_from_value():
    graph = init_graph_to_utils_test()['graph']
    p7 = nx.DiGraph()
    p7.add_node(1, label='x')
    p7.add_node(2)
    p7.add_node(3)
    p7.add_edge(1, 2, label='a')
    p7.add_edge(1, 3, label='a')

    embeddings = utils.get_embeddings(p7, graph)
    assert utils.get_key_from_value(embeddings[0], 1) == 6
    assert utils.get_key_from_value(embeddings[0], 2) == 8
    assert utils.get_key_from_value(embeddings[0], 3) == 1


def init_second_graph_to_utils_test():
    res = dict()
    g = nx.DiGraph()
    g.add_nodes_from(range(1, 9))
    g.add_edge(2, 1, label='a')
    g.add_edge(4, 1, label='a')
    g.add_edge(6, 1, label='a')
    g.add_edge(6, 8, label='a')
    g.add_edge(1, 3, label='b')
    g.add_edge(1, 5, label='b')
    g.add_edge(1, 7, label='b')
    g.nodes[1]['label'] = 'y'
    g.nodes[2]['label'] = 'x'
    g.nodes[3]['label'] = 'z'
    g.nodes[4]['label'] = 'x'
    g.nodes[5]['label'] = 'z'
    g.nodes[6]['label'] = 'x'
    g.nodes[7]['label'] = 'z'
    g.nodes[8]['label'] = 'w', 'x'
    res['g'] = g
    label_codes = LabelCodes(g)
    res['label_codes'] = label_codes
    p1 = nx.DiGraph()
    p1.add_node(1, label='x')
    p1.add_node(2)
    p1.add_edge(1, 2, label='a')
    res['p1'] = p1
    p2 = nx.DiGraph()
    p2.add_node(1, label='y')
    p2.add_node(2)
    p2.add_edge(1, 2, label='b')
    res['p2'] = p2
    ct = CodeTable(label_codes, g)
    ct.add_row(CodeTableRow(res['p1']))
    ct.add_row(CodeTableRow(res['p2']))
    ct.cover()
    res['ct'] = ct
    return res


def test_get_two_nodes_all_port():
    test = nx.DiGraph()
    test.add_node(1, is_Pattern=True)
    test.add_node(2, is_Pattern=True)
    test.add_node(3)
    test.add_node(4)
    test.add_node(5, is_Pattern=True)
    test.add_edge(1, 3)
    test.add_edge(1, 4)
    test.add_edge(2, 3)
    test.add_edge(2, 4)
    test.add_edge(5, 4)

    assert len(utils.get_two_nodes_all_port(1, 2, test)) == 2
    print('\n', utils.get_two_nodes_all_port(1, 2, test))
    assert len(utils.get_two_nodes_all_port(5, 2, test)) == 1
    print('\n', utils.get_two_nodes_all_port(5, 2, test))


def test_generate_candidates():
    res = init_second_graph_to_utils_test()
    ct = res['ct']

    candidates = utils.generate_candidates(ct.rewritten_graph(), ct)
    assert len(candidates) == 8
    assert candidates[0].first_pattern_label == 'P0'
    assert candidates[7].second_pattern_label == 'z'
    # print(candidates)


def test_compute_pattern_usage():
    res = init_second_graph_to_utils_test()
    ct = res['ct']
    """ {2: ['P0v2', 'P0v2', 'P0v2', 'P1v1', 'P1v1', 'P1v1'], 
    5: ['P0v1', 'P0v1'], 6: ['P0v2', 'wv1', 'xv1'], 
    9: ['P1v2', 'zv1'], 11: ['P1v2', 'zv1'], 
    13: ['P1v2', 'zv1']} """
    usage = utils.compute_pattern_usage(ct.rewritten_graph(), ['P0v2'], {2})
    assert usage == 3
    usage = utils.compute_pattern_usage(ct.rewritten_graph(), ['zv1'], {11, 9, 13})
    assert usage == 3

    rwg = nx.DiGraph()
    rwg.add_node(1, is_Pattern=True, label='P1')
    rwg.add_node(2)
    rwg.add_node(3, is_Pattern=True, label='P2')
    rwg.add_node(4, is_Pattern=True, label='P2')
    rwg.add_node(5)
    rwg.add_edge(1, 2, label='v1')
    rwg.add_edge(1, 5, label='v2')
    rwg.add_edge(3, 1, label='v2')
    rwg.add_edge(3, 5, label='v1')
    rwg.add_edge(4, 2, label='v1')
    usage = utils.compute_pattern_usage(rwg, ['P1v1', 'P1v2'], {2, 5})
    assert usage == 1
    # print(usage)


def test_compute_candidate_usage():
    res = init_second_graph_to_utils_test()
    ct = res['ct']
    candidates = utils.generate_candidates(ct.rewritten_graph(), ct)
    c1 = Candidate('P0', 'P0', [('v2', 'v2')])
    c1.first_pattern = res['p1']
    c1.second_pattern = res['p1']
    c1.data_port = {2}
    utils.compute_candidate_usage(ct.rewritten_graph(), c1, ct, candidates)
    assert c1.usage == 1

    c2 = Candidate('P0', 'P1', [('v2', 'v1')])
    c2.first_pattern = res['p1']
    c2.second_pattern = res['p2']
    c2.data_port = {2}
    utils.compute_candidate_usage(ct.rewritten_graph(), c2, ct, candidates)
    assert c2.usage == 3

    c3 = Candidate('z', 'P1', [('v1', 'v2')])
    c3.second_pattern = res['p2']
    with pytest.raises(ValueError):
        utils.compute_candidate_usage(ct.rewritten_graph(), c3, ct, candidates)

    c4 = Candidate('w', 'x', [('v1', 'v1')])
    c4.data_port = {6}
    utils.compute_candidate_usage(ct.rewritten_graph(), c4, ct, candidates)
    assert c4.usage == 1

    c5 = Candidate('P1', 'z', [('v2', 'v1')])
    c5.first_pattern = res['p2']
    c5.data_port = {11, 9, 13}
    utils.compute_candidate_usage(ct.rewritten_graph(), c5, ct, candidates)
    assert c5.usage == 3


def test_compute_pattern_embeddings():
    res = init_second_graph_to_utils_test()
    ct = res['ct']

    assert utils.compute_pattern_embeddings(ct.rewritten_graph(), 'P1') == 3
    assert utils.compute_pattern_embeddings(ct.rewritten_graph(), 'w') == 1


def test_is_candidate_port_exclusive():
    res = init_second_graph_to_utils_test()
    ct = res['ct']
    candidates = utils.generate_candidates(ct.rewritten_graph(), ct)
    c = Candidate('P0', 'P0', [('v1', 'v1')])
    c1 = Candidate('P0', 'P0', [('v2', 'v2')])
    utils.is_candidate_port_exclusive(candidates, c, 5)

    assert utils.is_candidate_port_exclusive(candidates, c, 5) is True
    assert utils.is_candidate_port_exclusive(candidates, c1, 2) is False


"""def test_get_candidates():
    res = init_graph2()
    ct = res['ct']
    restricted_candidates = utils.get_candidates(ct.rewritten_graph(), ct)
    assert len(restricted_candidates) == 8
    assert restricted_candidates[1].usage == 3.0
    assert restricted_candidates[6].exclusive_port_number == 3"""

""" Code Table """


def test_is_node_marked():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(1)
    test.nodes[1]['label'] = 'x', 'w'

    with pytest.raises(ValueError):
        ct.is_node_marked(6, test, 1, 'x')
        ct.is_node_marked(1, test, 1, 'a')

    assert ct.is_node_marked(1, test, 1, 'x') is False
    test.nodes[1]['cover_mark'] = {'x': 1}
    assert ct.is_node_marked(1, test, 1, 'x') is True


def test_is_node_labels_marked():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(1)
    test.nodes[1]['label'] = 'x', 'w'

    with pytest.raises(ValueError):
        ct.is_node_labels_marked(6, test, 1, 'x')
        ct.is_node_labels_marked(1, test, 1, 'a')

    test.nodes[1]['cover_mark'] = {'x': 1}
    assert ct.is_node_labels_marked(1, test, 1, ('x', 'w')) is False
    test.nodes[1]['cover_mark'] = {'x': 1, 'w': 1}
    assert ct.is_node_labels_marked(1, test, 1, ('x', 'w')) is True


def test_is_edge_marked():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(range(1, 3))
    test.add_edge(1, 2, label='e')

    assert ct.is_edge_marked(1, 2, test, 1, 'e') is False
    test[1][2]['cover_mark'] = {'e': 1}
    assert ct.is_edge_marked(1, 2, test, 1, 'e') is True


def test_mark_node():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(1, label='x')

    with pytest.raises(ValueError):
        ct.mark_node(1, test, 1, None)

    ct.mark_node(1, test, 1, 'x')
    assert ct.is_node_marked(1, test, 1, 'x') is True

    test.nodes[1]['label'] = test.nodes[1]['label'], 'w'
    ct.mark_node(1, test, 2, ('x', 'w'))

    assert ct.is_node_marked(1, test, 1, 'x') is False
    assert ct.is_node_marked(1, test, 2, 'x') is True
    assert ct.is_node_marked(1, test, 2, 'w') is True


def test_mark_edge():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(1, label='x')
    test.add_node(2)
    test.add_edge(1, 2, label='e')

    with pytest.raises(ValueError):
        ct.mark_edge(1, 2, test, 1, None)
        ct.mark_edge(1, 2, test, 1, 'f')

    ct.mark_edge(1, 2, test, 1, 'e')
    assert ct.is_edge_marked(1, 2, test, 1, 'e') is True

    ct.mark_edge(1, 2, test, 2, 'e')
    assert ct.is_edge_marked(1, 2, test, 1, 'e') is False
    assert ct.is_edge_marked(1, 2, test, 2, 'e') is True


def init_second_graph_to_code_table_test():
    gtest = nx.DiGraph()
    gtest.add_nodes_from(range(1, 6))
    gtest.add_edge(1, 2, label='e')
    gtest.add_edge(2, 3, label='e')
    gtest.add_edge(2, 4, label='e')
    gtest.add_edge(5, 2, label='e')
    gtest.nodes[1]['label'] = 'A'
    gtest.nodes[2]['label'] = 'A'
    gtest.nodes[3]['label'] = 'B'
    gtest.nodes[4]['label'] = 'B'
    gtest.nodes[5]['label'] = 'A'

    pattern = nx.DiGraph()
    pattern.add_nodes_from(range(1, 3))
    pattern.add_edge(1, 2, label='e')
    pattern.nodes[1]['label'] = 'A'
    embeddings = utils.get_embeddings(pattern, gtest)
    label_codes = LabelCodes(gtest)
    code_table = CodeTable(label_codes, gtest)
    rewritten_graph = nx.DiGraph()
    return gtest, pattern, embeddings, label_codes, code_table, rewritten_graph


def test_is_embedding_marked():
    gtest = init_second_graph_to_code_table_test()[0]
    pattern = init_second_graph_to_code_table_test()[1]
    embeddings = init_second_graph_to_code_table_test()[2]
    ct = init_second_graph_to_code_table_test()[4]
    assert ct.is_embedding_marked(embeddings[0], pattern, gtest, 1) is False

    gtest[1][2]['cover_mark'] = {'e': 1}
    assert ct.is_embedding_marked(embeddings[0], pattern, gtest, 1) is True

    ct.mark_edge(1, 2, gtest, 2, 'e')
    assert ct.is_embedding_marked(embeddings[0], pattern, gtest, 1) is False
    assert ct.is_embedding_marked(embeddings[0], pattern, gtest, 2) is True


def test_mark_embedding():
    gtest = init_second_graph_to_code_table_test()[0]
    pattern = init_second_graph_to_code_table_test()[1]
    embeddings = init_second_graph_to_code_table_test()[2]
    ct = init_second_graph_to_code_table_test()[4]
    ct.mark_embedding(embeddings[0], gtest, pattern, 1)

    assert ct.is_edge_marked(1, 2, gtest, 1, 'e') is True
    assert ct.is_node_marked(1, gtest, 1, 'A') is True
    assert ct.is_node_marked(2, gtest, 1, 'A') is False

    ct.mark_embedding(embeddings[0], gtest, pattern, 2)
    assert ct.is_edge_marked(1, 2, gtest, 1, 'e') is False
    assert ct.is_node_marked(1, gtest, 1, 'A') is False

    assert ct.is_edge_marked(1, 2, gtest, 2, 'e') is True
    assert ct.is_node_marked(1, gtest, 2, 'A') is True

    pattern.nodes[2]['label'] = 'A'
    ct.mark_embedding(embeddings[0], gtest, pattern, 1)
    assert ct.is_node_marked(2, gtest, 1, 'A') is True


def test_get_node_label_number():
    test = nx.Graph()
    test.add_node(1)
    ct = init_second_graph_to_code_table_test()[4]

    with pytest.raises(ValueError):
        ct.get_node_label_number(2, test)

    test.nodes[1]['label'] = 'a'
    assert ct.get_node_label_number(1, test) == 1

    test.nodes[1]['label'] = test.nodes[1]['label'], 'b'
    assert ct.get_node_label_number(1, test) == 2

    test.add_node(2)
    assert ct.get_node_label_number(2, test) == 0


def test_search_port():
    test1 = nx.DiGraph()
    test1.add_node(40, label='x')
    test1.add_node(41)
    test1.add_node(42)
    test1.add_node(43, label='y')
    test1.add_edge(40, 41, label='a')
    test1.add_edge(40, 42, label='a')
    test1.add_edge(40, 43, label='a')
    ptest1 = nx.DiGraph()
    ptest1.add_node(1, label='x')
    ptest1.add_node(2)
    ptest1.add_edge(1, 2, label='a')
    ct = init_second_graph_to_code_table_test()[4]

    embed = utils.get_embeddings(ptest1, test1)
    ct.mark_embedding(embed[0], test1, ptest1, 1)
    ct.mark_embedding(embed[1], test1, ptest1, 1)
    port_usage = dict()
    ports = ct.search_port(test1, embed[0], 1, ptest1, port_usage)
    assert ports[0][0] == 40
    assert ports[0][1] == 1
    assert (1 in port_usage.keys()) is True
    ct.search_port(test1, embed[1], 1, ptest1, port_usage)
    assert port_usage[1] != 1


def test_is_node_edges_marked():
    gtest = init_second_graph_to_code_table_test()[0]
    pattern = init_second_graph_to_code_table_test()[1]
    embeddings = init_second_graph_to_code_table_test()[2]
    ct = init_second_graph_to_code_table_test()[4]
    ct.mark_embedding(embeddings[0], gtest, pattern, 1)

    assert ct.is_node_edges_marked(gtest, 1, pattern, 1) is True
    assert ct.is_node_edges_marked(gtest, 2, pattern, 1) is False


def test_is_node_all_labels_marked():
    ct = init_second_graph_to_code_table_test()[4]
    test = nx.Graph()
    test.add_node(1, label='x')
    test.add_node(3, label='y')
    test.add_node(2)
    test.nodes()[2]['label'] = 'x', 'y'
    test.add_edge(1, 2, label='e')
    test.add_edge(2, 3, label='e')

    ct.mark_node(2, test, 1, 'x')
    ct.mark_node(1, test, 1, 'x')

    assert ct.is_node_all_labels_marked(2, test, 1) is False
    assert ct.is_node_all_labels_marked(1, test, 1) is True


def test_row_cover():
    gtest = init_second_graph_to_code_table_test()[0]
    pattern = init_second_graph_to_code_table_test()[1]
    embeddings = init_second_graph_to_code_table_test()[2]
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]

    ptest = nx.Graph()
    ptest.add_node(1, label='x')

    test = nx.Graph()
    test.add_node(1, label='x')
    test.add_node(3, label='y')
    test.add_node(2)
    test.nodes()[2]['label'] = 'x', 'y'
    test.add_edge(1, 2, label='e')
    test.add_edge(2, 3, label='e')

    row_test = CodeTableRow(ptest)
    row_test.set_embeddings(utils.get_embeddings(ptest, test))
    # with pytest.raises(ValueError):
    ct.row_cover(row_test, test, 1, rewritten_graph, 0)
    assert ct.is_node_marked(1, test, 1, ptest.nodes(data=True)[1]['label']) is True
    assert ct.is_node_marked(2, test, 1, ptest.nodes(data=True)[1]['label']) is True

    row = CodeTableRow(pattern)
    row.set_embeddings(embeddings)

    ct.row_cover(row, gtest, 1, rewritten_graph, 1)

    for edge in gtest.edges(data=True):
        assert ct.is_edge_marked(edge[0], edge[1], gtest, 1, gtest[edge[0]][edge[1]]['label']) is True

    for node in gtest.nodes(data=True):
        if node[1]['label'] == 'A':
            assert ct.is_node_marked(node[0], gtest, 1, node[1]['label']) is True
        else:
            assert ct.is_node_marked(node[0], gtest, 1, node[1]['label']) is False


def test_singleton_cover():
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]

    test = nx.Graph()
    test.add_node(1, label='x')
    test.add_node(3, label='y')
    test.add_node(2)
    test.nodes()[2]['label'] = 'x', 'y'
    test.add_edge(1, 2, label='e')
    test.add_edge(2, 3, label='e')
    res = ct.singleton_cover(test, 1, rewritten_graph)

    assert res[0]['x'] == 2
    assert res[0]['y'] == 2
    assert res[1]['e'] == 2


def test_rows():
    code_table = init_second_graph_to_code_table_test()[4]
    assert len(code_table.rows()) == 0


def test_add_row():
    pattern = init_second_graph_to_code_table_test()[1]
    code_table = init_second_graph_to_code_table_test()[4]

    row1 = CodeTableRow(pattern)

    pattern2 = nx.DiGraph()
    pattern2.add_node(1, label='B')

    row2 = CodeTableRow(pattern2)

    code_table.add_row(row1)
    code_table.add_row(row2)

    assert len(code_table.rows()[0].pattern().nodes()) == 2
    assert len(code_table.rows()[1].pattern().nodes()) == 1


def test_create_rewrite_edge():
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]
    rewritten_graph.add_node(1, label='40')
    rewritten_graph.add_node(2, label='P1')

    ct.create_rewrite_edge(rewritten_graph, 2, 40, pattern_port=1)
    assert ((2, 1) in list(rewritten_graph.edges(2))) is True
    assert rewritten_graph[2][1]['label'] == 'v1'

    ct.create_rewrite_edge(rewritten_graph, 2, 41, pattern_port=2)
    assert (3 in rewritten_graph.nodes()) is True
    assert rewritten_graph.nodes[3]['label'] == '41'
    assert ((2, 3) in list(rewritten_graph.edges(2))) is True
    assert rewritten_graph[2][3]['label'] == 'v2'


def test_create_vertex_singleton_node():
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]

    rewritten_graph.add_node(1, label='40')
    ct.create_vertex_singleton_node(rewritten_graph, 'x', 40)
    assert (2 in rewritten_graph.nodes()) is True
    assert ('is_Pattern' in rewritten_graph.nodes(data=True)[2]) is True
    assert rewritten_graph.nodes[2]['is_Pattern'] is True
    assert ('is_singleton' in rewritten_graph.nodes(data=True)[2]) is True
    assert rewritten_graph.nodes[2]['is_singleton'] is True
    assert rewritten_graph.nodes[2]['label'] == 'x'
    assert rewritten_graph[2][1]['label'] == 'v1'


def test_create_edge_singleton_node():
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]

    rewritten_graph.add_node(1, label='40')
    ct.create_edge_singleton_node(rewritten_graph, 'a', 40, 41)
    assert rewritten_graph.nodes[2]['is_Pattern'] is True
    assert rewritten_graph.nodes[2]['is_singleton'] is True
    assert rewritten_graph.nodes[3]['label'] == '41'
    assert rewritten_graph.nodes[2]['label'] == 'a'
    assert rewritten_graph[2][1]['label'] == 'v1'
    assert rewritten_graph[2][3]['label'] == 'v2'


def test_create_pattern_node():
    rewritten_graph = init_second_graph_to_code_table_test()[5]
    ct = init_second_graph_to_code_table_test()[4]

    rewritten_graph.add_node(1, label='40')
    ct.create_pattern_node(rewritten_graph, 1, [(40, 1)])
    assert rewritten_graph.nodes[2]['is_Pattern'] is True
    assert ('is_singleton' in rewritten_graph.nodes(data=True)[2]) is False
    assert rewritten_graph.nodes[2]['label'] == 'P1'
    assert rewritten_graph[2][1]['label'] == 'v1'


def init_graph_to_code_table_test():
    res = dict()
    graph = nx.DiGraph()
    graph.add_nodes_from(range(1, 9))
    graph.add_edge(2, 1, label='a')
    graph.add_edge(4, 1, label='a')
    graph.add_edge(6, 1, label='a')
    graph.add_edge(6, 8, label='a')
    graph.add_edge(8, 6, label='a')
    graph.add_edge(1, 3, label='b')
    graph.add_edge(1, 5, label='b')
    graph.add_edge(1, 7, label='b')
    graph.nodes[1]['label'] = 'y'
    graph.nodes[2]['label'] = 'x'
    graph.nodes[3]['label'] = 'z'
    graph.nodes[4]['label'] = 'x'
    graph.nodes[5]['label'] = 'z'
    graph.nodes[6]['label'] = 'x'
    graph.nodes[7]['label'] = 'z'
    graph.nodes[8]['label'] = 'w', 'x'

    res['graph'] = graph

    p1 = nx.DiGraph()
    p1.add_nodes_from(range(1, 4))
    p1.add_edge(1, 2, label='a')
    p1.add_edge(2, 3, label='b')
    p1.nodes[1]['label'] = 'x'
    p1.nodes[2]['label'] = 'y'
    p1.nodes[3]['label'] = 'z'
    row1 = CodeTableRow(p1)
    res['p1'] = p1
    res['row1'] = row1

    p2 = nx.DiGraph()
    p2.add_nodes_from(range(1, 3))
    p2.nodes[1]['label'] = 'x'
    p2.nodes[2]['label'] = 'x'
    p2.add_edge(1, 2, label='a')
    p2.add_edge(2, 1, label='a')
    row2 = CodeTableRow(p2)

    res['p2'] = p2
    res['row2'] = row2

    p3 = nx.DiGraph()
    p3.add_node(1, label='x')
    p3.add_node(2)
    p3.add_edge(1, 2, label='a')
    row3 = CodeTableRow(p3)
    res['p3'] = p3
    res['row3'] = row3

    p4 = nx.DiGraph()
    p4.add_node(1, label='z')
    p4.add_node(2)
    p4.add_edge(2, 1, label='b')
    row4 = CodeTableRow(p4)
    res['p4'] = p4
    res['row4'] = row4

    p5 = nx.DiGraph()
    p5.add_node(1, label='x')
    p5.add_node(2, label='y')
    p5.add_edge(1, 2, label='a')
    row5 = CodeTableRow(p5)
    res['p5'] = p5
    res['row5'] = row5

    p6 = nx.DiGraph()
    p6.add_node(1, label='y')
    p6.add_node(2, label='z')
    p6.add_edge(1, 2, label='b')
    row6 = CodeTableRow(p6)
    res['p6'] = p6
    res['row6'] = row6

    p7 = nx.DiGraph()
    p7.add_node(1, label='y')
    p7.add_node(2)
    p7.add_node(3)
    p7.add_edge(1, 2, label='a')
    p7.add_edge(1, 3, label='a')
    row7 = CodeTableRow(p7)
    res['p7'] = p7
    res['row7'] = row7
    lc = LabelCodes(graph)
    res['lc'] = lc
    ct = CodeTable(lc, graph)
    res['ct'] = ct
    return res


def test_cover():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    # ct.add_row(row7)
    ct.add_row(res['row3'])
    ct.add_row(res['row4'])
    ct.add_row(res['row5'])
    ct.add_row(res['row6'])
    ct.cover()

    # print(ct.display_ct())


"""def test_compute_ct_description_length():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    with pytest.raises(ValueError):
        ct.compute_ct_description_length()

    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 112.75

    ct.add_row(res['row1'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 113.55

    ct.add_row(res['row2'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 103.29

    ct.remove_row(res['row1'])
    ct.remove_row(res['row2'])

    ct.add_row(res['row3'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 102.23

    ct.add_row(res['row4'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 91.02

    ct.add_row(res['row5'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 108.91

    ct.add_row(res['row6'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 118.0"""


def test_rewritten_graph():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    ct.add_row(res['row5'])
    p8 = nx.DiGraph()
    p8.add_node(1)
    p8.add_node(2, label='z')
    p8.add_edge(1, 2, label='b')
    row8 = CodeTableRow(p8)
    ct.add_row(row8)
    ct.cover()
    print('\n data_port: ', ct.data_port())
    print('\n port count :', utils.count_port_node(ct.rewritten_graph()))
    # assert utils.count_port_node(ct.rewritten_graph()) == 6
    print('\n pattern infos :', utils.get_pattern_node_infos(ct.rewritten_graph()))
    # assert len(utils.get_pattern_node_infos(ct.rewritten_graph())['P0']) == 5
    print('\n port infos :', utils.get_port_node_infos(ct.rewritten_graph()))
    # can = utils.get_candidates(ct.rewritten_graph(), ct)
    # print(can)


"""def test_compute_rewritten_graph_description():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 143.51

    ct.add_row(res['row1'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 56.49

    ct.add_row(res['row2'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 38.05

    ct.remove_row(res['row1'])
    ct.remove_row(res['row2'])

    ct.add_row(res['row3'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 108.15

    ct.add_row(res['row4'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 72.84

    ct.add_row(res['row5'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 68.61

    ct.add_row(res['row6'])
    ct.cover()
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 68.61"""


def test_description_length_with_prequential_code():
    res = init_multi_graph()
    ct = res['ct']
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 291.22
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 446.34
    assert pytest.approx(ct.compute_rewritten_graph_description() + ct.description_length(), rel=1e-01) == 737.56
    ct.add_row(res['row1'])
    ct.add_row(res['row2'])
    ct.add_row(res['row3'])
    ct.cover()
    ct.compute_ct_description_length()
    assert pytest.approx(ct.description_length(), rel=1e-01) == 232.93
    assert pytest.approx(ct.compute_rewritten_graph_description(), rel=1e-01) == 108.92


def test_is_ct_edge_singleton():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    ct.add_row(res['row5'])
    p8 = nx.DiGraph()
    p8.add_node(1)
    p8.add_node(2, label='z')
    p8.add_edge(1, 2, label='b')
    row8 = CodeTableRow(p8)
    ct.add_row(row8)
    ct.cover()
    assert ct.is_ct_edge_singleton('a') is True
    assert ct.is_ct_edge_singleton('w') is False


def test_is_ct_vertex_singleton():
    res = init_graph_to_code_table_test()
    ct = res['ct']
    ct.add_row(res['row5'])
    p8 = nx.DiGraph()
    p8.add_node(1)
    p8.add_node(2, label='z')
    p8.add_edge(1, 2, label='b')
    row8 = CodeTableRow(p8)
    ct.add_row(row8)
    ct.cover()
    assert ct.is_ct_vertex_singleton('a') is False
    assert ct.is_ct_vertex_singleton('w') is True


""" Candidate"""


def test_set_usage():
    c = Candidate('P0', 'P1', [('v1', 'v2')])
    c.set_usage(2.0)

    assert c.usage == 2.0


def test_inverse():
    c = Candidate('P0', 'P1', [('v1', 'v2')])
    c_inverse = Candidate('P1', 'P0', [('v2', 'v1')])
    assert c.inverse() == c_inverse


def test_merge_candidate():
    pa = nx.DiGraph()
    pa.add_node(1, label='A')
    pa.add_node(2, label='B')
    pa.add_node(3, label='C')
    pa.add_edge(1, 2, label='a')
    pa.add_edge(2, 3, label='b')

    pb = nx.DiGraph()
    pb.add_node(1, label='D')
    pb.add_node(2, label='E')
    pb.add_node(3, label='F')
    pb.add_edge(1, 2, label='c')
    pb.add_edge(2, 3, label='d')

    c = Candidate('P0', 'P1', [('v2', 'v1')])
    c.first_pattern = pa
    c.second_pattern = pb

    graph = c.merge_candidate()

    assert len(graph.nodes()) == 5
    assert len(graph.nodes[2]['label']) == 2
    assert graph[2][4]['label'] == 'c'

    c1 = Candidate('P0', 'P0', [('v3', 'v2')])
    c1.first_pattern = pa
    c1.second_pattern = pa
    g1 = c1.merge_candidate()
    assert len(g1.nodes[3]['label']) == 2
    assert g1[4][3] is not None
    assert g1[3][5] is not None

    c2 = Candidate('P0', 'P1', [('v1', 'v1')])
    c2.first_pattern = pa
    c2.second_pattern = pb
    g2 = c2.merge_candidate()
    assert len(g2.nodes[1]['label']) == 2
    assert g2[1][4]['label'] == 'c'

    c3 = Candidate('P0', 'P1', [('v1', 'v1'), ('v3', 'v3')])
    c3.first_pattern = pa
    c3.second_pattern = pb
    g3 = c3.merge_candidate()
    assert len(g3.nodes()) == 4
    assert g3.nodes[1]['label'] == ('D', 'A')
    assert g3.nodes[3]['label'] == ('F', 'C')
    assert g3[4][3] is not None

    p1 = nx.DiGraph()
    p1.add_node(1, label='C')
    p1.add_node(2)
    p1.add_node(3, label='C')
    p1.add_edge(1, 2, label='single')
    p1.add_edge(3, 1, label='single')

    p2 = nx.DiGraph()
    p2.add_node(1, label='x')
    p2.add_node(2)
    p2.add_edge(1, 2, label='a')
    c4 = Candidate('P1', 'P1', [('v2', 'v3'), ('v3', 'v2'), ('v1', 'v1')])
    c4.first_pattern = p1
    c4.second_pattern = p1
    g4 = c4.merge_candidate()
    assert len(g4.nodes()) == 3


def test_final_pattern():
    pa = nx.DiGraph()
    pa.add_node(1, label='A')
    pa.add_node(2, label='B')
    pa.add_node(3, label='C')
    pa.add_edge(1, 2, label='a')
    pa.add_edge(2, 3, label='b')

    pb = nx.DiGraph()
    pb.add_node(1, label='D')
    pb.add_node(2, label='E')
    pb.add_node(3, label='F')
    pb.add_edge(1, 2, label='c')
    pb.add_edge(2, 3, label='d')
    c2 = Candidate('P0', 'P1', [('v1', 'v1')])
    c2.first_pattern = pa
    c2.second_pattern = pb
    g2 = c2.final_pattern()
    assert len(g2.nodes[1]['label']) == 2
    assert g2[1][4]['label'] == 'c'


""" GraphMDL test"""
g = nx.DiGraph()
g.add_nodes_from(range(1, 9))
g.add_edge(2, 1, label='a')
g.add_edge(4, 1, label='a')
g.add_edge(6, 1, label='a')
g.add_edge(6, 8, label='a')
g.add_edge(8, 6, label='a')
g.add_edge(1, 3, label='b')
g.add_edge(1, 5, label='b')
g.add_edge(1, 7, label='b')
g.nodes[1]['label'] = 'y'
g.nodes[2]['label'] = 'x'
g.nodes[3]['label'] = 'z'
g.nodes[4]['label'] = 'x'
g.nodes[5]['label'] = 'z'
g.nodes[6]['label'] = 'x'
g.nodes[7]['label'] = 'z'
g.nodes[8]['label'] = 'w', 'x'


def test_fit():
    with pytest.raises(ValueError):
        GraphMDL().fit(None)

    mdl = GraphMDL()
    mdl.fit(g)
    # mdl.summary()
    assert mdl.description_length() != 0.0

    assert GraphMDL().fit(g, timeout=0.01).description_length() != 0


def test_patterns():
    assert len(GraphMDL().fit(g).patterns()) == 3


def test_description_length():
    assert pytest.approx(GraphMDL().fit(g).description_length(), rel=1e-01) == 144.8


def test_initial_description_length():
    assert pytest.approx(GraphMDL().fit(g).initial_description_length(), rel=1e-01) == 256.3


def test_graphmdl_on_multidigraph():
    graph = init_multi_graph()['graph']
    GraphMDL().fit(graph)



from skmine.graph.graphmdl import utils
import math


class CodeTableRow:
    """
        Object to represent a row of the code table
        Its concerns only non-singleton pattern

        Parameters
        ----------
        pattern
            It's the row pattern structure
        pattern_usage:int ,default=None
            It's the row usage after a cover
        pattern_port_usage: dict ,default=None
            This is the usage of each port of the line pattern after the cover.
    """

    def __init__(self, pattern, pattern_usage=None, pattern_port_usage=None):
        self._pattern = pattern
        self._pattern_usage = pattern_usage
        self._pattern_port_usage = pattern_port_usage
        self._embeddings = []  # the pattern embeddings in the data
        self._code_length = 0.0
        self._port_code_length = None  # the pattern ports code
        self._description_length = 0.0
        self._used_embeddings = []  # used embeddings
        self.arrival_number = None

    def code_length(self):
        """ Provide the row code length
        Returns
        -------
        float
        """
        return self._code_length

    def port_code_length(self):
        """ Provide the row ports code length
        Returns
        -------
        dict
        """
        return self._port_code_length

    def pattern(self):
        """ Provide the row pattern
        Returns
        -------
        object
        """
        return self._pattern

    def pattern_usage(self):
        """ Provide the row pattern code
        Returns
        -------
        float
        """
        return self._pattern_usage

    def set_pattern_usage(self, pattern_code):
        """ Set the row pattern code
        Parameters
        ----------
        pattern_code
        """
        self._pattern_usage = pattern_code

    def pattern_port_usage(self):
        """ Provide the port code of the row pattern
        Returns
        -------
        dict
        """
        return self._pattern_port_usage

    def set_pattern_port_usage(self, port_code):
        """ set the port code of the row pattern
        Parameters
        ---------
        port_code
        """
        self._pattern_port_usage = port_code

    def set_embeddings(self, embeddings):
        """ Set the pattern row embeddings
        Parameters
        ----------
        embeddings
        """
        self._embeddings = embeddings

    def add_used_embeddings(self, embedding):
        self._used_embeddings.append(embedding)

    def used_embeddings(self):
        return self._used_embeddings

    def embeddings(self):
        """ Provide the pattern row embeddings
        Returns
        -------
        list
        """
        return self._embeddings

    def compute_code_length(self, rows_usage_sum):
        """ Compute the code length of the row and its ports
        Parameters
        ---------
        rows_usage_sum : total of usage for the code table rows
        """
        self._port_code_length = dict()

        # Compute pattern code length
        if self._pattern_usage == 0:
            self._code_length = 0.0
        else:
            self._code_length = utils.log2(self._pattern_usage, rows_usage_sum)

        # compute port usage sum
        port_usage_sum = 0.0
        for k in self._pattern_port_usage.keys():
            port_usage_sum = port_usage_sum + self._pattern_port_usage[k]

        # compute each port code length
        for p in self._pattern_port_usage.keys():
            if self._pattern_port_usage[p] == 0:
                self._port_code_length[p] = 0.0
            else:
                code = utils.log2(self._pattern_port_usage[p], port_usage_sum)
                if code == - 0.0:
                    self._port_code_length[p] = 0.0
                else:
                    self._port_code_length[p] = code

    def compute_description_length(self, standard_table):
        """ Compute the row  description length according kgmdl equation
        Parameters
        ---------
        standard_table
        """
        if self._pattern_usage is None:
            self._description_length = 0.0

        if self._pattern_port_usage is None or self._port_code_length is None:
            raise ValueError("Row's codes should be compute")
        self._description_length = 0.0

        code_port_total = 0.0
        for value in self._port_code_length.values():
            code_port_total += value

        port_desc = math.log2(len(self._pattern.nodes()) + 1)
        port_desc += math.log2(utils.binomial(len(self._pattern.nodes()), len(self._port_code_length)))
        # port_desc += code_port_total

        # self._description_length = self._code_length  # usage description
        self._description_length += utils.encode(self._pattern, standard_table)  # structure description
        self._description_length += port_desc  # ports description

    def description_length(self):
        """ Provide the row description length
        Returns
        -------
        float
        """
        return self._description_length

    def display_row(self):
        """ Display a row in a certain format
        Returns
        -------
        list
        """
        return [utils.draw_pattern(self._pattern), self._pattern_usage, self._code_length,
                len(self._pattern_port_usage),
                self._pattern_port_usage, self._port_code_length]

    def __str__(self):
        return "{} | {} |{} |{} |{} |{}" \
            .format(utils.display_graph(self._pattern), self._pattern_usage, self._code_length,
                    len(self._pattern_port_usage), self._pattern_port_usage,
                    self._port_code_length)



"""
The :mod:`skmine.periodic` implements tools to tackle periodicity in
datasets, preferably presentend in the form of event logs
"""
from .cycles import PeriodicCycleMiner, SingleEventCycleMiner
from .tree import PeriodicPatternMiner



"""SLIM pattern discovery"""

# Authors: Rémi Adon <remi.adon@gmail.com>
# License: BSD 3 clause

from collections import Counter, defaultdict
from functools import lru_cache, reduce
from itertools import chain

import numpy as np
import pandas as pd
from sortedcontainers import SortedDict
from pyroaring import BitMap as Bitmap

from ..base import BaseMiner, InteractiveMiner, MDLOptimizer
from ..utils import _check_D, supervised_to_unsupervised


def _to_vertical(D, stop_items=set(), return_len=False):
    res = defaultdict(Bitmap)
    idx = 0
    for idx, transaction in enumerate(D):
        for e in transaction:
            if e in stop_items:
                continue
            res[e].add(idx)
    if return_len:
        return dict(res), idx + 1
    return dict(res)


def _log2(values):
    res_index = values.index if isinstance(values, pd.Series) else None
    res = np.zeros(len(values), dtype=np.float32)
    res[values != 0] = np.log2(values[values != 0]).astype(np.float32)
    return pd.Series(res, index=res_index)


def cover(sct: dict, itemsets: list):
    """
    cover a standard codetable sct given itemsets

    Parameters
    ----------
    sct: dict[object, Bitmap]
        a standard codetable, i.e the vertical representation of a dataset
    itemsets: list[frozenset]
        itemsets from a given codetable

    Note
    ----
        sct is modified inplace
    """
    covers = dict()
    for iset in itemsets:
        it = [sct[i] for i in iset]
        usage = reduce(Bitmap.intersection, it).copy() if it else Bitmap()
        covers[iset] = usage
        for k in iset:
            sct[k] -= usage
    return covers


def generate_candidates_big(codetable, stack=set(), depth=None):
    """
    Generate candidates, but does not sort output by estimated gain

    The result is a python generator, not an in-memory list

    This results in slightly less accurate candidate generation,
    but avoids computing candidates that will never be evaluated,
    if coupled with an early stopping strategy.

    Parameters
    ----------
    codetable: SortedDict[frozenset, Bitmap]
        A codetable, sorted in Standard Candidate Order

    stack: set[frozenset], defaut=set()
        A stack of already seen itemsets, which will not be considered in output
        Note that this function updates the stack, passed as a reference

    See Also
    --------
    generate_candidates
    """
    assert isinstance(codetable, SortedDict)
    depth = depth or int(np.log2(len(codetable)) * 1e2)
    for idx, (X, X_usage) in enumerate(codetable.items()):
        Y = codetable.items()[idx + 1: idx + 1 + depth]
        _best_usage = 0
        best_XY = None
        for y, y_usage in Y:
            XY = X.union(y)
            if XY in stack:
                continue
            stack.add(XY)
            inter_len = y_usage.intersection_cardinality(X_usage)
            if inter_len > _best_usage:
                _best_usage = inter_len
                best_XY = XY

        if best_XY is not None:
            yield best_XY, _best_usage


def generate_candidates(codetable, stack=set()):
    """
    assumes codetable is sorted in Standard Candidate Order
    """
    return sorted(
        generate_candidates_big(codetable, stack=stack),
        key=lambda e: e[1],
        reverse=True,
    )


class SLIM(BaseMiner, MDLOptimizer, InteractiveMiner):
    """SLIM: Directly Mining Descriptive Patterns

    SLIM looks for a compressed representation of transactional data.
    This compressed representation if a set of descriptive patterns,
    and can be used to:

    - provide a natively interpretable modeling of this data
    - make predictions on new data, using this condensed representation as an encoding scheme


    Parameters
    ----------
    k: int, default=50
        Number of non-singleton itemsets to mine.
        A singleton is an itemset containing a single item.
    pruning: bool, default=True
        Either to activate pruning or not. Pruned itemsets may be useful at
        prediction time, so it is usually recommended to set it to `False`
        to build a classifier. The model will be less concise, but will lead
        to more accurate predictions on average.
    n_items: int, default=200
        Number of most frequent items to consider for mining.
        As SLIM is highly dependant from the set of symbols from which
        it refines its codetable,
        lowering this argument will significantly improve runtime.

        Note: The reconstruction is lossless from this set of items. If the input data
        has more than `n_items` items, then the reconstruction will be lossy w.r.t this
        input data.
    tol: float, default=0.5
        Minimum compression gain (in bits) for a candidate to be accepted


    Examples
    --------
    >>> from skmine.itemsets import SLIM
    >>> D = [['bananas', 'milk'], ['milk', 'bananas', 'cookies'], ['cookies', 'butter', 'tea']]
    >>> SLIM().fit(D).discover(singletons=True, usage_tids=True)
    (bananas, milk)    (0, 1)
    (butter, tea)         (2)
    (cookies,)         (1, 2)
    dtype: object

    References
    ----------
    .. [1]
        Smets, K & Vreeken, J
        "Slim: Directly Mining Descriptive Patterns", 2012

    .. [2] Gandhi, M & Vreeken, J
        "Slimmer, outsmarting Slim", 2014
    """

    def __init__(
            self, *, k=50, pruning=True, n_items=200, tol=0.5,
    ):
        self.n_items = n_items
        self.tol = tol
        self.standard_codetable_ = None
        self.codetable_ = SortedDict()
        self.model_size_ = None  # L(CT|D)
        self.data_size_ = None  # L(D|CT)
        self.pruning = pruning
        self.k = k

    def fit(self, D, y=None):  # pylint:disable = too-many-locals
        """fit SLIM on a transactional dataset

        This generates new candidate patterns and add those which improve compression,
        iteratibely refining ``self.codetable_``

        Parameters
        ----------
        D: iterable of iterables or array-like
            Transactional dataset, either as an iterable of iterables
            or encoded as tabular binary data
        """
        self.prefit(D, y=y)
        seen_cands = set()
        k = 0

        while k < self.k:
            candidates = self.generate_candidates(stack=seen_cands)
            for cand, _ in candidates:
                data_size, model_size, usages = self.evaluate(cand)
                diff = (self.model_size_ + self.data_size_) - (data_size + model_size)

                if diff >= self.tol:
                    self.update(
                        usages=usages, data_size=data_size, model_size=model_size
                    )

                    k = sum(map(lambda iset: len(iset) > 1, self.codetable_))
                if k >= self.k:
                    break

            if not candidates:  # if empty candidate generation
                Warning(f"could not find `{self.k}` itemsets, try with a lower `tol`")
                break

        return self

    def decision_function(self, D):
        """Compute covers on new data, and return code length

        This function is named ``decision_function`` because code lengths
        represent the distance between a point and the current codetable.

        Setting ``pruning`` to False when creating the model
        is recommended to cover unseen data, and especially when building a classifier.

        Parameters
        ----------
        D: pd.DataFrame or np.ndarray
            new data to make predictions on, in tabular format

        Example
        -------
        >>> from skmine.itemsets import SLIM; import pandas as pd
        >>> def to_tabular(D): return pd.Series(D).str.join('|').str.get_dummies(sep="|")
        >>> D = [['bananas', 'milk'], ['milk', 'bananas', 'cookies'], ['cookies', 'butter', 'tea']]
        >>> new_D = to_tabular([['cookies', 'butter']])
        >>> slim = SLIM().fit(to_tabular(D))
        >>> slim.decision_function(new_D)
        0   -1.321928
        dtype: float32

        See Also
        --------
        cover
        discover
        """
        mat = self.cover(D)
        code_lengths = self.discover(singletons=True, usage_tids=False)
        ct_codes = code_lengths / code_lengths.sum()
        codes = (mat * ct_codes).sum(axis=1).astype(np.float32)
        # positive sign on log2 to return negative distance : sklearn]
        r = _log2(codes)
        r[r == 0] = -np.inf  # zeros would fool a `shortest code wins` strategy
        return r

    def generate_candidates(self, stack=set()):
        """
        Generate candidates from the current codetable (SLIM is any-time)

        Note that `stack` is updated during the execution of this method.

        Parameters
        ----------
        stack: set[frozenset], default=None
            a stack of already-seen candidates to be excluded

        Returns
        -------
        iterator[tuple(frozenset, Bitmap)]
        """
        ct = SortedDict(self._standard_candidate_order, self.codetable_.items())
        return generate_candidates(ct, stack=stack)

    def evaluate(self, candidate):
        """
        Evaluate ``candidate``, considering the current codetable and a dataset ``D``

        Parameters
        ----------
        candidate: frozenset
            a new candidate to be evaluated

        Returns
        -------
        (float, float, dict)
            updated (data size, model size, codetable)
        """
        idx = self.codetable_.bisect(candidate)
        ct = list(self.codetable_)
        ct.insert(idx, candidate)
        D = {k: v.copy() for k, v in self.standard_codetable_.items()}
        CTc = cover(D, ct)

        decreased = set()
        for iset, usage in self.codetable_.items():  # TODO useless is size is too big
            if len(CTc[iset]) < len(usage):
                decreased.add(iset)

        data_size, model_size = self._compute_sizes(CTc)

        if self.pruning:
            CTc, data_size, model_size = self._prune(
                CTc, decreased, model_size, data_size
            )

        return data_size, model_size, CTc

    def update(self, candidate=None, model_size=None, data_size=None, usages=None):
        """
        Update the current codetable.

        If `candidate` is passed as None, `model_size`, `data_size` and `usages` will be used
        If `candidate` is not None, `model_size`, `data_size` and `usages`
        will be computed by calling `.evaluate`

        Parameters
        ----------
        candidate: frozenset, default=None
            candidate to be inserted

        model_size: float, default=None
            new model size (in bits) to be set

        data_size: float
            new data size (in bits) to be set

        usages: dict, default=None
            optional for usage outside this class
            eg. if one simply needs to include an itemset in the current codetable
            as in interactive data mining

        Raises
        ------
        AssertionError
        """
        assert not (candidate is None and usages is None)
        if usages is None:
            data_size, model_size, usages = self.evaluate(candidate)
        to_drop = {c for c in self.codetable_.keys() - usages.keys() if len(c) > 1}
        self.codetable_.update(usages)
        for iset in to_drop:
            del self.codetable_[iset]

        self.data_size_ = data_size
        self.model_size_ = model_size

    def cover(self, D):
        """
        cover unseen data

        items never seen are dropped out


        Examples
        --------
        >>> from skmine.itemsets import SLIM
        >>> D = ["ABC", "AB", "BCD"]
        >>> s = SLIM().fit(D)
        >>> s.cover(["BC", "AB"])
           (A, B)   (B,)   (C,)
        0   False   True   True
        1    True  False  False

        Returns
        -------
        pd.DataFrame
        """
        if hasattr(D, "shape") and len(D.shape) == 2:  # tabular
            D = _check_D(D)
            D_sct = {
                k: Bitmap(np.where(D[k])[0])
                for k in D.columns
                if k in self.standard_codetable_
            }
        else:  # transactional
            D_sct = _to_vertical(D)

        isets = self.discover(singletons=True, usage_tids=False)
        isets = isets[isets.index.map(set(D_sct).issuperset)]
        covers = cover(D_sct, isets.index)

        mat = np.zeros(shape=(len(D), len(covers)), dtype=bool)
        for idx, tids in enumerate(covers.values()):
            mat[tids, idx] = True
        return pd.DataFrame(mat, columns=list(covers.keys()))

    def discover(self, singletons=False, usage_tids=False, drop_null_usage=True):
        """Get a user-friendly copy of the codetable

        Parameters
        ----------
        singletons: bool, default=False
            Either to include itemsets of length 1 in the result
        usage_tids: bool, default=False
            Either returns the transaction ids for each itemset in tuple format or its codelength
        drop_null_usage: bool, default=True
            Either to include itemset with no usage in the training data
            (i.e itemsets under cover of other itemsets)

        Example
        -------
        >>> from skmine.itemsets import SLIM
        >>> D = ["ABC", "AB", "BCD"]
        >>> SLIM().fit(D).discover(singletons=True, usage_tids=True, drop_null_usage=False)
        (A, B)    (0, 1)
        (B,)         (2)
        (A,)          ()
        (C,)      (0, 2)
        (D,)         (2)
        dtype: object

        Returns
        -------
        pd.Series
            codetable containing patterns and ids of transactions in which they are used
        """
        s = {
            tuple(sorted(iset)): tids.copy()
            for iset, tids in self.codetable_.items()
            if len(tids) >= drop_null_usage and len(iset) > (not singletons)
        }
        s = pd.Series(list(s.values()), index=list(s.keys()))
        if not usage_tids:
            s = s.map(len).astype(np.uint32)
        return s

    def reconstruct(self):
        """reconstruct the original data from the current `self.codetable_`"""
        n_transactions = (
                max(map(Bitmap.max, filter(lambda e: e, self.codetable_.values()))) + 1
        )

        D = pd.Series([set()] * n_transactions)
        for itemset, tids in self.codetable_.items():
            D.iloc[list(tids)] = D.iloc[list(tids)].map(itemset.union)
        return D.map(sorted)

    @lru_cache(maxsize=1024)
    def get_support(self, *items):
        """
        Get support from an itemset

        Note
        ----
        Items in an itemset must be passed as positional arguments

        Unseen items will throw errors
        """
        a = items[-1]
        tids = self.standard_codetable_[a]
        if len(items) > 1:
            return tids & self.get_support(*items[:-1])
        return tids

    def _standard_cover_order(self, itemset):
        """
        Returns a tuple associated with an itemset,
        so that many itemsets can be sorted in Standard Cover Order
        """
        return -len(itemset), -len(self.get_support(*itemset)), tuple(itemset)

    def _standard_candidate_order(self, itemset):
        return -len(self.get_support(*itemset)), -len(itemset), tuple(itemset)

    def prefit(self, D, y=None):
        """
        Parameters
        ----------
        D: iterable of iterables or array-like
            Transactional dataset, either as an iterable of iterables
            or encoded as tabular binary data

        Note
        ----
        works in 3 steps

        1. ingest data `D`
        2. track bitmaps for the top `self.n_items` frequent items from `D`
        3. set `self.data_size_` and `self.model_size` given the standard codetable
        """
        if hasattr(D, "ndim") and D.ndim == 2:
            D = _check_D(D)
            if y is not None:
                D = supervised_to_unsupervised(D, y)  # SKLEARN_COMPAT
            item_to_tids = {k: Bitmap(np.where(D[k])[0]) for k in D.columns}
        else:
            item_to_tids = _to_vertical(D)
        sct = pd.Series(item_to_tids)
        usage = sct.map(len).astype(np.uint32)
        usage = usage.nlargest(self.n_items)
        sct = sct[usage.index]
        self.standard_codetable_ = sct

        ct_it = ((frozenset([e]), tids) for e, tids in sct.items())
        self.codetable_ = SortedDict(self._standard_cover_order, ct_it)

        codes = -_log2(usage / usage.sum())
        self._starting_codes = codes

        # L(code_ST(X)) = L(code_CT(X)), because CT=ST
        self.model_size_ = 2 * codes.sum()

        self.data_size_ = (codes * usage).sum()

        return self

    def _compute_sizes(self, codetable):
        """
        Compute sizes for both the data and the model

        .. math:: L(D|CT)
        .. math:: L(CT|D)

        Parameters
        ----------
        codetable : Mapping
            A series mapping itemsets to their usage tids

        Returns
        -------
        tuple(float, float)
            (data_size, model_size)
        """
        isets, usages = zip(
            *((_[0], len(_[1])) for _ in codetable.items() if len(_[1]) > 0)
        )
        usages = np.array(usages, dtype=np.uint32)
        codes = -_log2(usages / usages.sum())

        counts = Counter(chain(*isets))
        stand_codes_sum = sum(
            self._starting_codes[item] * ctr for item, ctr in counts.items()
        )

        model_size = stand_codes_sum + codes.sum()  # L(CTc|D) = L(X|ST) + L(X|CTc)
        data_size = (codes * usages).sum()
        return data_size, model_size

    def _prune(self, codetable, prune_set, model_size, data_size):
        """post prune a codetable considering itemsets for which usage has decreased

        Parameters
        ----------
        codetable: SortedDict
        prune_set: set
            itemsets in ``codetable`` for which usage has decreased
        model_size: float
            current model_size for ``codetable``
        data_size: float
            current data size when encoding ``D`` with ``codetable``

        Returns
        -------
        new_codetable, new_data_size, new_model_size: SortedDict, float, float
            a tuple containing the pruned codetable, and new model size and data size
            w.r.t this new codetable
        """
        prune_set = {k for k in prune_set if len(k) > 1}  # remove singletons
        while prune_set:
            cand = min(prune_set, key=lambda e: len(codetable[e]))
            prune_set.discard(cand)

            ct = list(codetable)
            ct.remove(cand)

            D = {
                k: v.copy() for k, v in self.standard_codetable_.items()
            }  # TODO avoid data copies
            CTp = cover(D, ct)
            decreased = {
                k for k, v in CTp.items() if len(k) > 1 and len(v) < len(codetable[k])
            }

            d_size, m_size = self._compute_sizes(CTp)

            if d_size + m_size < model_size + data_size:
                codetable.update(CTp)
                del codetable[cand]
                prune_set.update(decreased)
                data_size, model_size = d_size, m_size

        return codetable, data_size, model_size



"""
LCM: Linear time Closed item set Miner
as described in `http://lig-membres.imag.fr/termier/HLCM/hlcm.pdf`
"""

# Authors: Rémi Adon <remi.adon@gmail.com>
#          Luis Galárraga <galarraga@luisgalarraga.de>
#          Hermann Courteille <hermann.courteille@irisa.fr>
#          Cyril Regan <cyril.regan@irisa.fr>
#          Thomas Betton <thomas.betton@irisa.fr>
#
# License: BSD 3 clause

from collections import defaultdict
from itertools import takewhile
import pandas as pd
import os
import shutil
from joblib import Parallel, delayed
from sortedcontainers import SortedDict
from pyroaring import BitMap as Bitmap

from ..utils import _check_min_supp
from ..utils import filter_maximal

from ..base import BaseMiner, DiscovererMixin


class LCM(BaseMiner, DiscovererMixin):
    """
    Linear time Closed item set Miner.

    LCM can be used as a **generic purpose** miner, yielding some patterns
    that will be later submitted to a custom acceptance criterion.

    It can also be used to simply discover the set of **closed itemsets** from
    a transactional dataset.

    Parameters
    ----------
    min_supp: int or float, default=0.2
        The minimum support for itemsets to be rendered in the output
        Either an int representing the absolute support, or a float for relative support
        Default to 0.2 (20%)

    n_jobs : int, default=1 The number of jobs to use for the computation. Each single item is attributed a job to
        discover potential itemsets, considering this item as a root in the search space. **Processes are preferred**
        over threads. **Carefully adjust the number of jobs** otherwise the results may be corrupted especially if you
        have the following warning: UserWarning: A worker stopped while some jobs were given to the executor.

    References
    ----------
    .. [1]
        Takeaki Uno, Masashi Kiyomi, Hiroki Arimura
        "LCM ver. 2: Efficient mining algorithms for frequent/closed/maximal itemsets", 2004

    .. [2] Alexandre Termier
        "Pattern mining rock: more, faster, better"

    Examples
    --------

    >>> from skmine.itemsets import LCM
    >>> from skmine.datasets.fimi import fetch_chess
    >>> chess = fetch_chess()
    >>> lcm = LCM(min_supp=2000)
    >>> patterns = lcm.fit_discover(chess)
    >>> patterns.head()
        itemset support
    0      [58]    3195
    1      [52]    3185
    2  [58, 52]    3184
    3      [29]    3181
    4  [58, 29]    3180
    >>> patterns[patterns.itemset.map(len) > 3]  # doctest: +SKIP
    """

    def __init__(self, *, min_supp=0.2, n_jobs=1, verbose=False):
        _check_min_supp(min_supp)
        self.min_supp = min_supp  # cf docstring: minimum support provided by user
        self.max_length = -1  # maximum length of an itemset
        self.verbose = verbose
        self._min_supp = _check_min_supp(self.min_supp)
        self.item_to_tids = SortedDict()  # Dict : key item ordered by decreasing frequency , value : tids for this
        # item
        self.ord_item_freq = []  # list of ordered item by decreasing frequency
        self.n_jobs = n_jobs  # number of jobs launched by joblib
        self.lexicographic_order = False  # if true, the items of each itemset are returned in lexicographical order
        self.return_tids = False

    def fit(self, D, y=None):
        """
        fit LCM on the transactional database, by keeping records of singular items
        and their transaction ids.

        Parameters
        ----------
        D: pd.Series or iterable
            a transactional database. All entries in this D should be lists.
            If D is a pandas.Series, then `(D.map(type) == list).all()` should return `True`

        Raises
        ------
        TypeError
            if any entry in D is not iterable itself OR if any item is not **hashable**
            OR if all items are not **comparable** with each other.
        """
        n_transactions_ = 0
        item_to_tids_ = defaultdict(Bitmap)
        for transaction in D:
            for item in transaction:
                item_to_tids_[item].add(n_transactions_)
            n_transactions_ += 1

        if isinstance(self.min_supp, float):  # make support absolute if needed
            self._min_supp = self.min_supp * n_transactions_

        low_supp_items = [k for k, v in item_to_tids_.items() if len(v) < self._min_supp]
        for item in low_supp_items:  # drop low freq items
            del item_to_tids_[item]

        ord_freq_list = sorted(item_to_tids_.items(), key=lambda item: len(item[1]), reverse=True)
        ord_freq_dic = defaultdict(Bitmap)
        ord_item_freq = []
        for idx, element in enumerate(ord_freq_list):
            item, tid = element
            ord_item_freq.append(item)
            ord_freq_dic[idx] = tid  # rename most frequent item like cat by 0, second  dog by 1

        self.item_to_tids = SortedDict(ord_freq_dic)  # {0:tids0, 1:tids1 ....}
        self.ord_item_freq = ord_item_freq  # [cat, dog, '0', ...]

        return self  # for call .fit().discover() in DiscovererMixin()

    def discover(self, *, return_tids=False, lexicographic_order=False, max_length=-1, out=None):
        """Return the set of closed itemsets, with respect to the minimum support

        Parameters
        ----------
        D : pd.Series or Iterable
            The input transactional database
            Where every entry contain singular items
            Items must be both hashable and comparable

        return_tids: bool, default=False
            Either to return transaction ids along with itemset.
            Default to False, will return supports instead

        lexicographic_order: bool, default=False
            Either the order of the items in each itemset is not ordered or the items are ordered lexicographically

        max_length: int, default=-1 Maximum length of an itemset. By default, -1 means that LCM returns itemsets of
        all lengths.

        out : str, default=None
            File where results are written. Discover return None. The 'out' option is usefull 
            to save memory : Instead of store all branch of lcm-tree in memory , each root 
            branch of lcm is written in a separated file in dir (TEMP_dir), and all files are
            concatenated in the final 'out' file.

        Returns
        -------
        pd.DataFrame
            DataFrame with the following columns
                ==========  =================================
                itemset     a `list` of co-occured items
                support     frequence for this itemset
                ==========  =================================

            if `return_tids=True` then
                ==========  =================================
                itemset     a `tuple` of co-occured items
                support     frequence for this itemset
                tids        a bitmap tracking positions
                ==========  =================================

        Example
        -------
        >>> from skmine.itemsets import LCM
        >>> D = [[1, 2, 3, 4, 5, 6], [2, 3, 5], [2, 5]]
        >>> LCM(min_supp=2).fit_discover(D, lexicographic_order=True)
             itemset support
        0     [2, 5]       3
        1  [2, 3, 5]       2
        >>> LCM(min_supp=2).fit_discover(D, return_tids=True)
             itemset support       tids
        0     [2, 5]       3  (0, 1, 2)
        1  [2, 5, 3]       2     (0, 1)
        """
        self.lexicographic_order = lexicographic_order
        self.return_tids = return_tids
        self.max_length = max_length

        if out is None:  # store results in memory
            dfs = Parallel(n_jobs=self.n_jobs, prefer="processes")(
                delayed(self._explore_root)(item, tids, root_file=None) for item, tids in
                list(self.item_to_tids.items())
            )  # dfs is a list of dataframe
            # make sure we have something to concat
            columns = ["itemset", "support"] if not self.return_tids else ["itemset", "support", "tids"]
            dfs.append(pd.DataFrame(columns=columns))
            df = pd.concat(dfs, axis=0, ignore_index=True)

            return df

        else:  # store results in files
            temp_dir = 'TEMP_dir'  # temporary dir where root items branch files are written
            if os.path.exists(temp_dir):  # remove dir TEMP_dir if it exists
                shutil.rmtree(temp_dir)
            os.mkdir(temp_dir)  # create dir TEMP_dir

            Parallel(n_jobs=self.n_jobs, prefer="processes")(
                delayed(self._explore_root)(
                    item, tids, root_file=f"{temp_dir}/root{k}.dat") for k, (
                    item, tids) in enumerate(list(self.item_to_tids.items()))
            )

            with open(out, 'w') as outfile:  # concatenate all itemsroot files located in temp_dir in a single file
                for fname in [f"{temp_dir}/root{k}.dat" for k in
                              range(len(list(self.item_to_tids.items())))]:  # all items root files
                    with open(fname) as infile:
                        for line in infile:
                            if line.strip():  # to skip empty lines
                                outfile.write(line)
            shutil.rmtree(temp_dir)  # remove the temporary dir where root files are written

            return None

    def _explore_root(self, item, tids, root_file=None):
        it = self._inner((frozenset(), tids), item)
        columns = ["itemset", "support"] if not self.return_tids else ["itemset", "support", "tids"]
        df = pd.DataFrame(data=it, columns=columns)

        if self.verbose and not df.empty:
            print("LCM found {} new itemsets from root item : {}".format(len(df), item))

        if root_file is not None:  # for writing the items root files in dir self.temp_dir

            if os.path.exists(root_file):  # delete the root file if it already exists
                os.remove(root_file)

            self.write_df_tofile(root_file, df)
            return None
        else:
            return df

    def _inner(self, p_tids, limit):

        p, tids = p_tids
        # project and reduce DB w.r.t P
        cp = (
            item
            for item, ids in reversed(self.item_to_tids.items())
            if tids.issubset(ids)
            if item not in p
        )

        # items are in reverse order, so the first consumed is the max
        max_k = next(takewhile(lambda e: e >= limit, cp), None)

        if max_k is not None and max_k == limit:
            p_prime = (p | set(cp) | {max_k})  # max_k has been consumed when calling next()
            # sorted items in ouput for better reproducibility
            itemset = [self.ord_item_freq[ind] for ind in list(p_prime)]
            itemset = sorted(itemset) if self.lexicographic_order else itemset

            if len(itemset) <= self.max_length or self.max_length == -1:
                if not self.return_tids:
                    yield itemset, len(tids)
                else:
                    yield itemset, len(tids), tids

            candidates = self.item_to_tids.keys() - p_prime
            candidates = candidates[: candidates.bisect_left(limit)]
            for new_limit in candidates:
                ids = self.item_to_tids[new_limit]
                if tids.intersection_cardinality(ids) >= self._min_supp:
                    # new pattern and its associated tids
                    new_p_tids = (p_prime, tids.intersection(ids))
                    yield from self._inner(new_p_tids, new_limit)

    def write_df_tofile(self, filename, df):
        with open(filename, 'w') as fw:  # write the items root files
            for index, row in df.iterrows():
                fw.write(f"({row['support']}) {' '.join(map(str, row['itemset']))}\n")
                if self.return_tids:
                    fw.write(f"{' '.join(map(str, row['tids']))}\n")


class LCMMax(LCM):
    """
    Linear time Closed item set Miner adapted to Maximal itemsets (or borders).

    A maximal itemset is an itemset with no frequent superset.

    Parameters
    ----------
    min_supp: int or float, default=0.2
        The minimum support for itemsets to be rendered in the output
        Either an int representing the absolute support, or a float for relative support
        Default to 0.2 (20%)

    n_jobs : int, default=1
        The number of jobs to use for the computation. Each single item is attributed a job
        to discover potential itemsets, considering this item as a root in the search space.
        **Processes are preferred** over threads.

    See Also
    --------
    LCM
    """

    def _inner(self, p_tids, limit):
        p, tids = p_tids
        # project and reduce DB w.r.t P
        cp = (
            item
            for item, ids in reversed(self.item_to_tids.items())
            if tids.issubset(ids)
            if item not in p
        )
        max_k = next(cp, None)  # items are in reverse order, so the first consumed is the max

        if max_k is not None and max_k == limit:
            p_prime = (p | set(cp) | {max_k})  # max_k has been consumed when calling next()
            candidates = self.item_to_tids.keys() - p_prime
            candidates = candidates[: candidates.bisect_left(limit)]
            no_cand = True

            for new_limit in candidates:
                ids = self.item_to_tids[new_limit]
                if tids.intersection_cardinality(ids) >= self._min_supp:
                    no_cand = False
                    # get new pattern and its associated tids
                    new_p_tids = (p_prime, tids.intersection(ids))
                    yield from self._inner(new_p_tids, new_limit)

            if no_cand:  # only if no child node. This is how we PRE-check for maximality
                itemset = set([self.ord_item_freq[ind] for ind in p_prime])

                if not self.return_tids:
                    yield itemset, len(tids)
                else:
                    yield itemset, len(tids), tids

    def discover(self, *args, **kwargs):
        outfile = kwargs.get('out')
        kwargs['out'] = None
        patterns = super().discover(**kwargs)
        # keep only maximal itemsets
        maximals = filter_maximal(patterns["itemset"])
        patterns = patterns[patterns.itemset.isin(maximals)].copy()

        # keeps only itemsets smaller than max_length
        if self.max_length != -1:
            length_mask = patterns.itemset.apply(lambda x: len(x) <= self.max_length)
            patterns = patterns[length_mask]
            patterns.reset_index(drop=True, inplace=True)

        # return a list of itemset sorted or not
        patterns.loc[:, "itemset"] = patterns["itemset"].map(
            lambda i: sorted(list(i)) if self.lexicographic_order else list(i))

        if outfile:
            self.write_df_tofile(outfile, patterns)
            return None
        else:
            return patterns

    setattr(discover, "__doc__", LCM.discover.__doc__.replace("closed", "maximal"))
    setattr(discover, "__doc__", LCM.discover.__doc__.split("Example")[0])



"""
scikit-mine
Pattern mining algorithms for the Python ecosystem
"""

# TODO : define logger
# see : https://github.com/joblib/joblib/issues/634#issuecomment-367321025
# joblib currently disable logging inside delayed functions
__author__ = """scikit-mine team"""
__email__ = """remi.adon@gmail.com"""
__version__ = "0.0.8"  # pylint: disable= undefined-variable, pointless-statement
__credits__ = "Centre Inria Rennes - Bretagne Atlantique"



"""
Base IO for all FIMI datasets
All datasets are available here : `http://fimi.uantwerpen.be/data/`
"""
import os
import wget

import pandas as pd

from ._base import get_data_home

BASE_URL = "http://fimi.uantwerpen.be/data/"


def _read_dat(filepath, int_values=True, separator=' '):
    """Read a local dataset file whose separator can be customized and whose values are either integers or strings.

    Parameters
    ----------
    filepath : str
        Indicate the path of the file to be read

    int_values : bool
        Specify if the items in the file are all integers. If not, then the items are considered as strings.

    separator : str
        Specify a separator between items other than the default space

    Returns
    -------
    list
        Return a transaction list composed for each transaction of a list of items
    """
    with open(filepath, 'r') as f:
        try:
            lines = f.read().splitlines()
        except UnicodeDecodeError:
            print(f"The file {filepath} is already present in your data_home but it is a binary file, it must be "
                  f"deleted.")
            raise

    transactions = []
    for line in lines:
        tmp = line.rstrip()
        tmp = list(map(int, tmp.split(separator))) if int_values else tmp.split(separator)
        transactions.append(tmp)

    return transactions


def fetch_file(filepath, separator=' '):
    """Loader for files in FIMI format

    Parameters
    ----------
    filepath : str
        Path of the file to load

    separator : str
        Indicate a custom separator between the items. By default, it is a space.

    Returns
    -------
    pd.Series
        Transactions from the requested dataset,
        as an in-memory pandas Series
    """
    transactions = _read_dat(filepath, int_values=False, separator=separator)
    s = pd.Series(transactions, name=os.path.splitext(os.path.basename(filepath))[0])

    return s


def fetch_any(filename, data_home=None):
    """Base loader for all datasets from the FIMI repository
    Each unique transaction will be represented as a Python list in the resulting pandas Series

    see: http://fimi.uantwerpen.be/data/

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `~/scikit_mine_data/` subfolders.

    filename : str
        Name of the file to fetch

    Returns
    -------
    pd.Series
        Transactions from the requested dataset,
        as an in-memory pandas Series
    """
    data_home = data_home or get_data_home()
    filepath = os.path.join(data_home, filename)
    name, _ = os.path.splitext(filename)
    if filename in os.listdir(data_home):  # already fetched
        it = _read_dat(filepath)
    else:  # not fetched yet
        url = BASE_URL + filename
        wget.download(url, filepath)
        it = _read_dat(filepath)

    s = pd.Series(it, name=name)
    return s


def fetch_chess(data_home=None):
    """Fetch and return the chess dataset (Frequent Itemset Mining)

    ====================   ==============
    Nb of items                        75
    Nb of transactions               3196
    Avg transaction size             37.0
    Density                         0.493
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the chess dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("chess.dat", data_home=data_home)


def fetch_connect(data_home=None):
    """Fetch and return the connect dataset (Frequent Itemset Mining).

    ====================   ==============
    Nb of items                       129
    Nb of transactions              67557
    Avg transaction size             43.0
    Density                         0.333
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the connect dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("connect.dat", data_home=data_home)


def fetch_mushroom(data_home=None, return_mush_y=False):
    """Fetch and return the mushroom dataset (Frequent Itemset Mining)

    The Mushroom data set includes descriptions of hypothetical samples corresponding
    to 23 species of gilled mushrooms in the Agaricus and Lepiota Family.

    It contains information about 8124 mushrooms (transactions).
    4208 (51.8%) are edible and 3916 (48.2%) are poisonous.

    The data contains 22 nomoinal features plus the class attribure (edible or not).
    These features were translated into 117 items.

    ====================   ==============
    Nb of items                       117
    Nb of transactions               8124
    Avg transaction size             22.0
    Density                         0.188
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    return_mush_y: bool, default=False.
        If True, returns a tuple for both the data and the associated labels
        (0 for edible, 1 for poisonous)

    Returns
    -------
    mush : pd.Series
        Transactions from the mushroom dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.

    (mush, y) : tuple
        if ``return_D_y`` is True

    Examples
    --------
    >>> from skmine.datasets.fimi import fetch_mushroom
    >>> from skmine.datasets.utils import describe
    >>> mush, y = fetch_mushroom(return_mush_y=True)
    >>> describe(mush)['n_items']
    119
    >>> y.value_counts()
    0    4208
    1    3916
    Name: mushroom, dtype: int64
    """
    mush = fetch_any("mushroom.dat", data_home=data_home)
    if return_mush_y:
        y = mush.str[0]
        y = y.replace(2, 0)  # 2 is edible, 1 is poisonous
        return mush, y

    return mush


def fetch_pumsb(data_home=None):
    """Fetch and return the pumsb dataset (Frequent Itemset Mining)

    The Pumsb dataset contains census data for population and housing.

    ====================   ==============
    Nb of items                      2113
    Nb of transactions              49046
    Avg transaction size             74.0
    Density                         0.035
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the pumsb dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("pumsb.dat", data_home=data_home)


def fetch_pumsb_star(data_home=None):
    """Fetch and return the pumsb_star dataset (Frequent Itemset Mining)

    ====================   ==============
    Nb of items                      2088
    Nb of transactions              49046
    Avg transaction size            50.48
    Density                         0.024
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the pumsb_star dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("pumsb_star.dat", data_home=data_home)


def fetch_kosarak(data_home=None):
    """Fetch and return the kosarak dataset (Frequent Itemset Mining)

    Click-stream data from a hungarian on-line news portal.

    ====================   ==============
    Nb of items                     36855
    Nb of transactions             990002
    Avg transaction size              8.1
    Density                      0.000220
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the kosarak dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("kosarak.dat", data_home=data_home)


def fetch_retail(data_home=None):
    """Fetch and return the retail dataset (Frequent Itemset Mining)

    Contains market basket data from a Belgian retail store, anonymized.

    see: http://fimi.uantwerpen.be/data/retail.pdf

    ====================   ==============
    Nb of items                     16470
    Nb of transactions              88162
    Avg transaction size             10.3
    Densisty                     0.000626
    ====================   ==============

    Retail market basket data set supplied by a anonymous Belgian retail supermarket store.

    Results in approximately 5 months of data.
    The total amount of receipts being collected equals 88,163.

    In total, 5,133 customers have purchased at least one product during the data collection period

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default,
        all scikit-mine data is stored in `scikit-mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the retail dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.
    """
    return fetch_any("retail.dat", data_home=data_home)


def fetch_accidents(data_home=None):
    """Fetch and return the accidents dataset (Frequent Itemset Mining)

    Traffic accident data, anonymized.

    see: http://fimi.uantwerpen.be/data/accidents.pdf

    ====================   ==============
    Nb of items                       468
    Nb of transactions             340183
    Avg transaction size           33.807
    Density                         0.072
    ====================   ==============

    Parameters
    ----------
    data_home : optional, default: None
        Specify another download and cache folder for the datasets. By default
        all scikit-mine data is stored in `~/scikit_mine_data`.

    Returns
    -------
    pd.Series
        Transactions from the accidents dataset, as an in-memory pandas Series.
        Each unique transaction is represented as a Python list.

    """
    return fetch_any("accidents.dat", data_home=data_home)



import copy
import logging
import time
import networkx
from skmine.base import BaseMiner
from skmine.graph.graphmdl.label_codes import LabelCodes
from skmine.graph.graphmdl.code_table import CodeTable
from skmine.graph.graphmdl.code_table_row import CodeTableRow
from skmine.graph.graphmdl import utils


def _order_pruning_rows(row):
    """ Sort the pruning row by their usage"""
    return row.pattern_usage()


class GraphMDL(BaseMiner):
    """
        This is a python re-implementation of the GraphMDL family of approaches for extracting a small and
        descriptive set of graph patterns from graph data.
        This re-implementation supports directed graphs only, but supports multi-graphs.

        This is a python re-implementation of the original Java algorithm, which is available at 'https://gitlab.inria.fr/fbariatt/graphmdl'.
        As such, some functionalities are not available in this version, such as a full support of graph automorphisms.

        Author: Arnauld Djedjemel

        References
        ----------
         F. Bariatti, "Mining Tractable Sets of Graph Patterns with the Minimum Description Length Principle", PhD thesis, Université de Rennes 1, 2021. Available: https://hal.inria.fr/tel-03523742

         F. Bariatti, P. Cellier, and S. Ferré. "GraphMDL+ : interleaving the generation and MDL-based selection of graph patterns", in Proceedings of the 36th Annual ACM Symposium on Applied Computing, Mar. 2021, pp. 355–363. doi: 10.1145/3412841.3441917.


    """

    def __init__(self):
        self._data = None
        self._label_codes = None
        self._code_table = None
        self._rewritten_graph = None
        self._description_length = 0.0
        self._patterns = set()
        self._already_test = []  # list of the candidates already tested
        self._pruning_rows = []  # list of the rows who can prune
        self._old_usage = dict()  # Mapper for the rows old usage, useful for the pruning
        self._timeout = None
        self._initial_description_length = 0

    def _init_graph_mdl(self):
        """
            Initialize the GraphMDL elements such as the label code,
            the initial code table (CT0), and run a cover computation to create the first rewritten graph.
        """
        self._already_test = []
        self._label_codes = LabelCodes(self._data)  # label codes creation
        # CT0 creation
        self._code_table = CodeTable(self._label_codes, self._data)
        # CT0 cover
        self._code_table.cover()
        self._rewritten_graph = self._code_table.rewritten_graph()
        self._description_length = self._code_table.compute_total_description_length()
        self._initial_description_length = self._description_length
        utils.MyLogger.info(f"\n initial CT \n {self._code_table}")
        utils.MyLogger.info("GraphMDL+ run ...")
        utils.MyLogger.info(f"Initial description length = {round(self._description_length, 2)}")

    def fit(self, D, timeout=None):
        """
            Execute GraphMDl on a given data graph

            Parameters
            ----------
            D : networkx graph. All edges need to be labeled.
            timeout: int , default=None
                Maximum time for the algorithm execution (approx.).

            Returns
            -------
            GraphMDL
        """
        if timeout is not None and timeout != 0:
            self._timeout = timeout
        if D is None:
            raise ValueError("You should give a graph")
        else:
            self._data = D
            self._init_graph_mdl()
            self._anytime_graph_mdl_with_timeout()
            return self

    def _anytime_graph_mdl_with_timeout(self):
        """ Anytime graph mdl with timeout
        Returns
        --------
        GraphMDL
        """
        begin = time.time()
        current = 0
        stop = False
        while not stop:
            if self._timeout is not None:
                if current > self._timeout:
                    stop = True
                    break
            utils.MyLogger.info("Candidate generation and sort start .....")
            b = time.time()
            candidates = utils.generate_candidates(self._rewritten_graph, self._code_table)
            candidates.sort(reverse=True, key=self._order_candidates)
            utils.MyLogger.info(f"Candidate generation and sort end ..........time ={time.time() - b}")
            utils.MyLogger.info(f"candidates number {len(candidates)}")
            utils.MyLogger.info("GraphMDL best Ct search start .....")
            if self._timeout is not None:
                current = time.time() - begin
                if self._stop_by_time(current, self._timeout):
                    break
            if len(candidates) != 0:
                b = time.time()
                for candidate in candidates:
                    if candidate not in self._already_test:
                        # Add a candidate to a ct, cover and compute description length
                        if self._timeout is not None:
                            current = time.time() - begin
                            if self._stop_by_time(current, self._timeout):
                                break
                        self._compute_old_usage()
                        row = CodeTableRow(candidate.final_pattern())
                        self._code_table.add_row(row)
                        self._code_table.cover()
                        temp_code_length = self._code_table.compute_total_description_length()
                        self._already_test.append(candidate)
                        # if the new ct is better than the old, break and generate new candidates
                        # with the new ct
                        if temp_code_length < self._description_length:
                            self._rewritten_graph = self._code_table.rewritten_graph()
                            self._description_length = temp_code_length
                            utils.MyLogger.info("New DL ", self._description_length)
                            utils.MyLogger.info(f"new pattern added: {utils.display_graph(row.pattern())}")
                            utils.MyLogger.info(f"search time = {time.time() - b}")
                            self._compute_pruning_candidates()
                            self._pruning()
                            break
                        elif temp_code_length > self._description_length and candidates.index(candidate) == len(
                                candidates) - 1:
                            utils.MyLogger.info("None best code table found")
                            utils.MyLogger.info(f"search time = {time.time() - b}")
                            self._code_table.remove_row(row)
                            stop = self._graph_mdl_end()
                            if self._timeout is not None:
                                current = self._timeout
                            break
                        else:
                            # if the candidate not improve the result, remove it to the code table
                            self._code_table.remove_row(row)
                    else:
                        utils.MyLogger().debug("Already test")
            else:
                stop = self._graph_mdl_end()
                if self._timeout is not None:
                    current = self._timeout
        return self

    def _stop_by_time(self, passed_time, timeout):
        """
            Check if the passed time surpasses the timeout

            Parameters
            ---------
            passed_time : int
                The passed time
            timeout : int
                The maximum time

            Returns
            --------
            bool
        """
        if passed_time >= timeout:
            return self._graph_mdl_end()
        else:
            return False

    def _graph_mdl_end(self):
        """
            Complete the graph mdl algorithm by cover the code table and stop

            Returns
            -------
            bool
        """
        utils.MyLogger.info("GraphMDL+ end .....")
        self._code_table.cover()
        self._rewritten_graph = self._code_table.rewritten_graph()
        self._description_length = self._code_table.compute_total_description_length()
        utils.MyLogger.info(f"Final description length = {round(self._description_length, 2)}")
        utils.MyLogger.info(f"Number of patterns found = {len(self.patterns())}")

        return True

    def _order_candidates(self, candidate):
        """
            Provide the candidate elements to order candidates
            Parameters
            ----------
            candidate

            Returns
            -------
            list
        """
        return [candidate.usage, candidate.exclusive_port_number,
                -self._label_codes.encode(candidate.final_pattern())]

    def _compute_old_usage(self):
        """
            Store patterns usage before compute new code table patterns usage
        """
        self._old_usage = dict()
        for r in self._code_table.rows():
            self._old_usage[r.pattern()] = r.pattern_usage()

    def _compute_pruning_candidates(self):
        """
            Find the row where their usage has decreased since the last usage,
            it's the step before the code table pruning
        """
        for r in self._code_table.rows():
            if r.pattern() in self._old_usage.keys():
                if r.pattern_usage() < self._old_usage[r.pattern()]:
                    self._pruning_rows.append(r)

    def _pruning(self):
        """
            Make the code table pruning as krimp pruning.

            That's consist of remove row in code table who are unnecessary,
            because without them the code table is better.
        """
        utils.MyLogger.info(f"Pruning start ....")
        self._compute_old_usage()  # compute old pattern usage
        self._pruning_rows.sort(key=_order_pruning_rows)  # sort the pruning rows
        for r in self._pruning_rows:
            self._code_table.remove_row(r)
            self._code_table.cover()
            if self._code_table.compute_total_description_length() < self._description_length:
                self._rewritten_graph = self._code_table.rewritten_graph()
                self._description_length = self._code_table.compute_total_description_length()
                self._pruning_rows.remove(r)
                self._compute_pruning_candidates()  # recompute the pruning candidates
                self._pruning()
            else:
                self._code_table.add_row(r)

        utils.MyLogger.info("Pruning end ....")

    def patterns(self):
        """
            Return the patterns found by the algorithm after fit has been called.

            Returns
            -------
            set
        """
        self._patterns = set()
        if self._code_table is not None:
            for r in self._code_table.rows():
                if r.code_length() != 0:
                    self._patterns.add(r.pattern())

            for s in self._code_table.singleton_code_length().keys():
                self._patterns.add(utils.create_singleton_pattern(s, self._code_table))

            return self._patterns

        else:
            return ValueError("The fit method must be called first")

    def description_length(self):
        """
            Return the MDL description length (model + encoded data) for the best code table found by the algorithm.

            Returns
            --------
            float
        """
        return self._description_length

    def initial_description_length(self):
        """
            Return the MDL description length for the initial --singleton-only-- code table CT0.

            Returns
            -------
            float
        """
        return self._initial_description_length

    def discover(self, *args, **kwargs):
        """
            Provide a summary of the algorithm execution
        """
        if self._code_table is not None:
            print(self._code_table.display_ct())
            print("final description length : ", self._description_length)
            print("Non singleton patterns found : ")
            for p in self.patterns():
                print('\n', utils.display_graph(p))
            if len(self._code_table.singleton_code_length()) != 0:
                print("Singleton patterns found : ")
                for s in self._code_table.singleton_code_length().keys():
                    print("\n", s)
        else:
            raise ValueError("The fit method must be called first")




"""
SLIM vectorizer, using SLIM as a feature extraction scheme
"""

import numpy as np
import pandas as pd

from ..base import TransformerMixin, _get_tags
from ..itemsets import SLIM
from ..itemsets.slim import _to_vertical, cover

STRATEGIES = ("codes", "one-hot")


def _filter_stop_items(D, stop_items):
    for t in D:
        yield set(t).difference(stop_items)


class SLIMVectorizer(SLIM, TransformerMixin):
    """SLIM mining, turned into a feature extraction step for sklearn

    `k` new itemsets (associations of one or more items) are learned at training time

    The model (pattern set) is then used to cover new data, in order of usage.
    This is similar to one-hot-encoding, except the dimension will be much more concise,
    because the columns will be patterns learned via an MDL criterion.

    Parameters
    ----------
    strategy: str, default="codes"
        If the chosen strategy is set to `one-hot`, non-zero cells are filled with ones.

        If the chosen `strategy` is left to `codes`, non-zero cells are filled with code lengths,
        i.e the probabity of the pattern in the training data.

    k: int, default=5
        Number of non-singleton itemsets to mine.
        A singleton is an itemset containing a single item.

        Calls to `.transform` will output pandas.DataFrame with `k` columns

    pruning: bool, default=False
        Either to activate pruning or not.

    stop_items: iterable, default=None
        Set of items to filter out while ingesting the input data.


    Examples
    --------
    >>> from skmine.feature_extraction import SLIMVectorizer
    >>> D = [['bananas', 'milk'], ['milk', 'bananas', 'cookies'], ['cookies', 'butter', 'tea']]
    >>> SLIMVectorizer(k=2).fit_transform(D)
       (bananas, milk)  (cookies,)
    0              0.4         0.0
    1              0.4         0.4
    2              0.0         0.4


    See Also
    --------
    skmine.itemsets.SLIM

    Notes
    -----
    This transformer does not output scipy.sparse matrices,
    as SLIM should learn a concise description of the data,
    and covering new data with this small set of high usage
    patterns should output matrices with very few zeros.
    """

    def __init__(
        self, strategy="codes", *, k=5, pruning=False, stop_items=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.k = k
        self.pruning = pruning
        self.stop_items = stop_items
        if strategy not in STRATEGIES:
            raise ValueError(f"strategy must be one of {STRATEGIES}")
        self.strategy = strategy

    def fit(self, D, y=None):
        if self.stop_items is not None:
            D = _filter_stop_items(D, stop_items=self.stop_items)
        return super().fit(D)

    def transform(self, D, y=None):
        """Transform new data

        Parameters
        ----------
        D: iterable
            transactional data

        Returns
        -------
        pd.DataFrame
            a dataframe of `len(D)` rows and `self.k` columns

        See Also
        --------
        skmine.itemsets.SLIM.cover
        """
        stop_items = self.stop_items or set()
        D_sct, _len = _to_vertical(D, stop_items=stop_items, return_len=True)

        code_lengths = self.discover(
            usage_tids=False, singletons=True, drop_null_usage=False
        )
        code_lengths = code_lengths[code_lengths.index.map(set(D_sct).issuperset)]

        isets = code_lengths.nlargest(self.k)  # real usages sorted in decreasing order
        covers = cover(D_sct, isets.index)

        mat = np.zeros(shape=(_len, len(covers)))
        for idx, tids in enumerate(covers.values()):
            mat[tids, idx] = 1
        mat = pd.DataFrame(mat, columns=list(covers.keys()))

        if self.strategy == "codes":
            ct_codes = code_lengths[isets.index] / code_lengths.sum()
            mat = (mat * ct_codes).astype(np.float32)

        return mat

    def _get_tags(self):
        return {**_get_tags(self), **{"stateless": True}}  # modfies output shape



"""
Callback API for scikit-mine
"""
from inspect import signature, getsource

import re


def has_self_assigment(f):
    """
    Parameters
    ----------
    f: callable

    check if no assignement is made on the ``self`` keyword
    """
    try:
        p = r"self(\.\w+)?\s*=.*"
        return any(re.finditer(p, getsource(f)))
    except TypeError:
        return False


def _get_params(fn):
    assert callable(fn)
    try:
        sig = signature(fn)
        params = list(sig.parameters)
    except ValueError:
        params = list()
    return params


def post(self, func_name, callback):
    """decorator, call callback on returned values, after main function"""
    func = getattr(self, func_name)
    assert callable(func)
    callback_params = _get_params(callback)

    def _(*args, **kwargs):
        res = func(*args, **kwargs)
        if "self" in callback_params:
            _w = (res,) if len(callback_params) == 2 else res
            args_ = (self,) + (_w if len(callback_params) > 1 else tuple())
        else:
            args_ = res
        try:
            callback(*args_)
        except TypeError:
            callback(args_)  # list.extend
        return res

    return _


class CallBacks(dict):
    """
    A collection of callbacks

    Works by defining functions to be called after the execution of the function they target

    Parameters
    ----------
    key-value pairs
        Keys must be string, values must be callables

    Examples
    --------
    >>> class A():
    ...     def f(self):
    ...         return 10
    ...
    >>> from skmine.callbacks import CallBacks
    >>> stack = list()
    >>> callbacks = CallBacks(f=stack.append)
    >>> a = A()
    >>> callbacks(a)
    >>> a.f()
    10
    >>> stack
    [10]
    """

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)
        self._check()

    def _check(self):
        # TODO : inspect source code from callbacks and check no assigment on inner self
        for v in self.values():
            if not callable(v):
                raise TypeError(f"values must be callables, found {type(v)}")
            if has_self_assigment(v):  # TODO : only allow lambdas or builtins ?
                raise ValueError("callbacks should not modify `self` attributes")

    def _frozen(self, *args, **kwargs):
        raise NotImplementedError(f"{type(self)} is immutable")

    __setitem__ = _frozen
    update = _frozen

    def __call__(self, miner):
        miner_methods = [
            f_name for f_name in dir(miner) if callable(getattr(miner, f_name))
        ]
        for f_name in self.keys():
            if not f_name in miner_methods:
                raise ValueError(
                    f"{f_name} found in callbacks while there is not corresponding function"
                )

        for callback_name, callback in self.items():
            new_meth = post(miner, callback_name, callback)
            # TODO : lock assignement to "once for all"
            # re-executing a cell in a notebook can lead to callbacks being called many times
            setattr(miner, callback_name, new_meth)


def _print_positive_gain(self, data_size, model_size, *_):
    diff = (self.model_size_ + self.data_size_) - (data_size + model_size)
    if diff > 0.01:
        print(
            "data size : {:.2f} | model size : {:.2f}".format(data_size, model_size)
        )

def _print_candidates_size(self, candidates):
    print("{} new candidates considered".format(len(candidates)))


mdl_prints = CallBacks(
    evaluate=_print_positive_gain, generate_candidates=_print_candidates_size
)
mdl_prints.__doc__ = """
Base callback for miners which inherit the :class:`skmine.base.MDLOptimizer`

Prints data size and model size when compression has improved,
only if ``verbose`` is set to True for the miner to attach.

Examples
--------
>>> from skmine.callbacks import mdl_prints
>>> from skmine.base import MDLOptimizer
>>> class MyMDLMiner(MDLOptimizer):
...     def __init__(self):
...         self.codetable_ = dict()
...     def generate_candidates(self):
...         return [(2,), (2, 3), (2, 4)]
...     def evaluate(self):
...         pass

>>> miner = MyMDLMiner()
>>> mdl_prints(miner)
>>> miner.generate_candidates()
3 new candidates considered
[(2,), (2, 3), (2, 4)]
"""



"""Periodic pattern mining with a MDL criterion"""
# Authors: Rémi Adon <remi.adon@gmail.com>
#          Esther Galbrun <esther.galbrun@inria.fr>
#
# License: BSD 3 clause

import warnings
from itertools import groupby

import numpy as np
import pandas as pd

from pyroaring import BitMap as Bitmap

from ..base import BaseMiner, DiscovererMixin, MDLOptimizer
from ..utils import intersect2d, sliding_window_view

log = np.log2

INDEX_TYPES = (
    pd.DatetimeIndex,
    pd.RangeIndex,
    pd.Int64Index,
)


def residual_length(S_alpha, n_occs_tot, dS):
    """
    compute L(o) = L(t) + L(a) for all (a, t) in S_alpha
    i.e the length from a block of residual events

    Parameters
    ----------
    S_a: np.ndarray of shape or scalar
        array containing indices for the event `alpha` to consider

    n_occs_tot: int
        total number of occurrences in the original data

    dS: int
        max - min from original events
    """
    card = (
        S_alpha.shape[0] if isinstance(S_alpha, np.ndarray) else 1
    )  # TODO : remove me
    return log(dS + 1) - log(card / float(n_occs_tot))


def cycle_length(S, inter, n_event_tot, dS):
    """
    Parameters
    ----------
    S : np.array of type int64
        a collection of cycles, all having the same length : r
        The width of S is then r

    inter: np.array of type int64
        a collection of inter occurrence gaps, all having the same length: r - 1

    n_event_tot: int
        number of events in the original events

    dS: int
        max - min from original events

    Returns
    -------
    tuple:
        lengths for (a, r, p, tau, E)
    """
    r = S.shape[1]
    assert inter.shape[1] == r - 1  # check inter occurrences compliant with events
    p = np.median(inter, axis=1)
    E = inter - p.reshape((-1, 1))
    dE = E.sum(axis=1)
    S_size = len(S) + r - 1

    L_a = -log(S_size / n_event_tot)  # FIXME
    L_r = log(S_size)
    L_p = log(np.floor((dS - dE) / (r - 1)))
    L_tau = log(dS - dE - (r - 1) * p + 1)
    L_E = 2 * E.shape[1] + np.abs(E).sum(axis=1)

    return L_a, L_r, L_p, L_tau, L_E


def compute_cycles_dyn(S, n_tot, max_length=100):
    """
    Parameters
    ----------
    S: pd.Index or np.array
        a Series of occurrences
    n_tot: int
        total number of occurrences in the original events
    """
    _, cut_points = get_table_dyn(S, n_tot, max_length)
    splits = _recover_splits_rec(cut_points, 0, len(S) - 1)

    cycles = list()
    covered = Bitmap()
    for length, g in groupby(splits, key=lambda e: e[1] - e[0]):
        g = list(g)
        if length >= 2:  # eq to length + 1 >= 3
            curr_cycles = np.vstack([S[s : e + 1] for s, e in g])  # COPY
            cycles.append(curr_cycles)
            for s, e in g:
                covered.update(range(s, e + 1))

    return list(reversed(cycles)), covered


def get_table_dyn(S: pd.Index, n_tot: int, max_length=100):
    """
    Parameters
    ----------
    S: pd.Index or np.ndarray
        a Series of occurrences
    n_tot: int
        total number of occurrences in the original events
    max_length: int, default=None
        maximum number of occurrences for a cycle to cover,
        by default it will be set to :math:`\log_{2}\left(|S|\right)`

    """
    diffs = np.diff(S)
    triples = sliding_window_view(S, 3)
    diff_pairs = sliding_window_view(diffs, 2)
    dS = S.max() - S.min()

    score_one = residual_length(1, n_tot, dS)  # 1 really ?

    scores = sum(cycle_length(triples, diff_pairs, len(S), dS))
    change = scores > 3 * score_one
    scores[change] = 3 * score_one  # inplace replacement
    cut_points = np.array([-1] * len(scores), dtype=object)
    cut_points[~change] = None

    scores = dict(zip(((i, i + 2) for i in range(len(scores))), scores))
    cut_points = dict(zip(scores.keys(), cut_points))

    max_length = min([len(S), max_length])
    for k in range(4, max_length + 1):
        w = sliding_window_view(S, k)
        _diffs = sliding_window_view(diffs, k - 1)
        _s = sum(cycle_length(w, _diffs, len(S), dS))

        for ia, best_score in enumerate(_s):
            cut_point = None
            iz = ia + k - 1
            for im in range(ia, iz):
                if im - ia + 1 < 3:
                    score_left = score_one * (im - ia + 1)
                else:
                    score_left = scores[(ia, im)]
                if iz - im < 3:
                    score_right = score_one * (iz - im)
                else:
                    score_right = scores[(im + 1, iz)]

                if score_left + score_right < best_score:
                    best_score = score_left + score_right
                    cut_point = im
            scores[(ia, iz)] = best_score
            cut_points[(ia, iz)] = cut_point

    return scores, cut_points


def extract_triples(S, l_max=None):
    """
    Extract cycles of length 3 given a list of occurrences S
    Parameters
    ----------
    S: pd.Index
        input occurrences

    l_max: float
        maximum absolute difference between two occurrences to considered
        for inclusion in the same triple.
        By default it will be set to the median of the
        inter-occurrence differences from S.
    """
    triples = list()
    l_max = l_max or np.median(np.diff(S))

    # TODO : precompute diffs instead of for loop inner computation

    for idx, occ in enumerate(S[1:-1], 1):
        righties = S[idx + 1 :]
        lefties = S[:idx]
        righties_diffs = righties - occ
        lefties_diffs = lefties - occ
        dists = np.array(np.meshgrid(lefties_diffs, righties_diffs)).T.reshape(-1, 2)
        grid = np.abs(dists)
        keep = np.abs(grid[:, 0] - grid[:, 1]) < l_max
        t = occ + dists[keep]
        if t.size != 0:
            e = np.array([t[:, 0], np.array([occ] * t.shape[0]), t[:, 1]]).T
            assert np.issubdtype(e.dtype, np.number) and e.shape[1] == 3
            triples.append(e)
    if triples:
        return np.vstack(triples)
    return None


def merge_triples(triples, n_merge=10):
    """
    Parameters
    ----------
    triples: ndarray
        cycles of size 3 (i.e triples.shape[1] == 3)
    n_merge: int
        maximum number of merge operation to perform

    Returns
    -------
    list[np.ndarray]
        a list of cycles
    """
    if triples is None:
        return list()
    res = [triples]
    for idx in range(1, n_merge + 1):
        prev = res[idx - 1]
        lefties = prev[:, :2]
        righties = prev[:, -2:]
        _, left_idx, right_idx = intersect2d(lefties, righties, return_indices=True)
        if left_idx.size == 0 or right_idx.size == 0:
            break
        merged = np.hstack([prev[right_idx], prev[left_idx, 2:]])
        res.append(merged)
        to_delete = np.union1d(left_idx, right_idx)
        res[idx - 1] = np.delete(res[idx - 1], to_delete, axis=0)
        if (np.unique(merged) == np.unique(merged)).all():
            res.pop(idx - 1)
            break
    return list(reversed(res))  # inverse order of length


def _recover_splits_rec(cut_points, ia, iz):
    if (ia, iz) in cut_points:
        if cut_points[(ia, iz)] is None:
            return [(ia, iz)]
        im = cut_points[(ia, iz)]
        if im >= 0:
            return _recover_splits_rec(cut_points, ia, im) + _recover_splits_rec(
                cut_points, im + 1, iz
            )
    return []


def _remove_zeros(numbers: pd.Series):
    n = 0
    while (numbers % 10 == 0).all():
        numbers //= 10
        n += 1
    return numbers, n


def _reconstruct(start, period, dE):
    """
    Reconstruct occurrences,
    starting from `start`, and
    correcting `period` with a delta for all deltas in `dE`,
    `len(dE)` occurrences are reconstructed

    Parameters
    ----------
    start: int or datetime
        starting point for the event
    period: int or timedelta
        period between two occurrences
    d_E: np.array of [int|timedelta]
        inters occurrences deltas
    """
    occurrences = [start]
    current = start
    for d_e in dE:
        e = current + period + d_e
        occurrences.append(e)
        current = e
    return occurrences


def _generate_candidates_batch(
    S_a: np.array, n_occs_tot: int, max_length=100, presort=False
):
    """
    Generate occurrences to be merged together for form a cycle
    The result if a list of np.array, sorted by decreasing order of width
    """
    if len(S_a) < 3:
        return list()
    if presort:
        S_a = np.sort(S_a)
    dS = S_a[-1] - S_a[0]
    l_max = np.log2(dS + 1) - 2
    cycles, covered = compute_cycles_dyn(S_a, n_occs_tot, max_length)
    covered = Bitmap(covered)

    if len(S_a) - len(covered) > 3:  # add triples if necessary
        _all = Bitmap(range(len(S_a)))
        _S_a = S_a[_all - covered]
        triples = extract_triples(_S_a, l_max)
        merged = merge_triples(triples)
        cycles.extend(merged)

    return list(sorted(cycles, key=lambda _: _.shape[1], reverse=True))


def generate_candidates(S_a, n_occs_tot, max_length=100):
    """
    Returns
    -------
    A list of candidate batches. Each batch contains candidates occurrences of the same length,
    hence stored in a common `numpy.ndarray`.
    Batches are sorted by decreasing order of width, so that we consider larger candidates first.
    """
    occ_batches = _generate_candidates_batch(S_a, n_occs_tot, max_length=max_length)
    res = list()
    for cand_batch in occ_batches:
        length = cand_batch.shape[1]
        E = np.diff(cand_batch, axis=1)
        period = np.floor(np.median(E, axis=1)).astype("int64")
        dE = (E.T - period).T
        tids = [
            Bitmap(_)
            for _ in np.searchsorted(S_a, cand_batch.reshape(-1)).reshape(
                cand_batch.shape
            )
        ]

        mdl_cost = sum(cycle_length(cand_batch, E, n_occs_tot, S_a[-1] - S_a[0]))
        df = pd.DataFrame(
            dict(
                start=cand_batch[:, 0],
                length=length,
                period=period,
                dE=dE.tolist(),
                tids=tids,
                cost=mdl_cost,
            )
        )
        res.append(df)

    res = pd.concat(res, ignore_index=True) if res else pd.DataFrame()
    return res


def evaluate(cands: pd.DataFrame, k: int):
    """
    Evaluate candidates `cands`, given S

    Unlike the original implementation by Galbrun & Al.,
    if an occurrence is present in more than one candidate cycle, we keep the cycle
    with the greatest length.

    Parameters
    ----------
    cands: pd.DataFrame
        a dataframe containing the candidates, as well-formed cycles
    k: int
        number of cycles to keep for the same occurence covered
    """
    idx = cands.explode("tids").groupby("tids").cost.nsmallest(k).index
    if isinstance(idx, pd.MultiIndex):
        idx = idx.get_level_values(1)
    idx = idx.unique()
    return cands.loc[idx]


class SingleEventCycleMiner(BaseMiner, DiscovererMixin):
    """
    Miner periodic cycles, but in the scope of a single event

    To this end, we only accept 1 dimensional arrays as input
    PeriodicCycleMiner will operate by instantiating one SingleEventCycleMiner per event.

    Parameters
    ----------
    max_length: int, default=20
        maximum length for a cycle
    keep_residuals: bool, default=False
        either to keep residuals (lossless compression) or not (lossy compression)
    n_occs_tot: int, default=None
        Number of total occurrences. When instanciating this class on a subset of the original
        data, this can be useful to pass the number of occurrences in this original data.
    k: int, default=2
        Maximum number of cycles covering the same occurrence
    """

    def __init__(self, max_length=20, keep_residuals=False, n_occs_tot=None, k=2):
        self.cycles_ = pd.DataFrame(columns=["start", "length", "period", "dE"])
        self.residuals_ = list()
        self.max_length = max_length
        self.keep_residuals = keep_residuals
        if n_occs_tot:
            assert isinstance(n_occs_tot, int)
        self.n_occs_tot = n_occs_tot
        self._dS = None  # difference between the last occurrence and the first
        self.k = k
        self.tid_pad = 0
        # TODO : set self.n_zeros_ at the scale of an event-projected miner like this one ?

    def fit(self, X):
        """
        Parameters
        ----------
        X: np.array of shape (n_occs, )
            Input occurrences

        Notes
        -----
        X will be reshaped to a 1-d vector in any case
        """
        X = np.array(X)  # COPY
        X = np.sort(X.reshape(-1))
        n_occs_tot = self.n_occs_tot or len(X)
        candidates = generate_candidates(X, n_occs_tot, max_length=self.max_length)

        if candidates.empty:
            warnings.warn(
                """could not generate candidates for the input sequence,
                the model is left empty"""
            )
            if self.keep_residuals:
                self.residuals_ = X
            return self

        cycles = evaluate(candidates, self.k)
        self.cycles_ = cycles
        if self.keep_residuals:
            # careful with `~`
            uncovered = Bitmap(range(len(X))) - Bitmap.union(*self.cycles_.tids)
            self.residuals_ = X[uncovered]

        # cycles.dE = cycles.dE.map(np.array) ?
        return self

    def discover(self):
        """
        Expose cycles as a brand new dataframe
        """
        return self.cycles_.copy(deep=True)

    evaluate = evaluate

    def reconstruct(self):
        """Reconstruct the original occurrences from the current cycles.
        Residuals will also be included, as the compression scheme is lossless

        Denoting as :math:`\sigma(E)` the sum of the shift corrections for a cycle
        :math:`C`, we have

        .. math::
            \Delta(C)=(r-1) p+\sigma(E)

        Returns
        -------
        pd.Series
            The reconstructed dataset

        Notes
        -----
        The index of the resulting pd.Series will not be sorted
        """
        alpha_occurrences = list()
        for start, period, dE in self.cycles_[["start", "period", "dE"]].values:
            occurrences = _reconstruct(start, period, dE)
            alpha_occurrences.extend(occurrences)
        alpha_occurrences.extend(self.residuals_)
        return np.array(alpha_occurrences)


class PeriodicCycleMiner(BaseMiner, MDLOptimizer, DiscovererMixin):
    """
    Mining periodic cycles with a MDL Criterion

    PeriodicCycleMiner is an approach to mine periodic cycles from event logs
    while relying on a Minimum Description Length (MDL) criterion to evaluate
    candidate cycles. The goal here is to extract a set of cycles that characterizes
    the periodic structure present in the data

    A cycle is defined a 5-tuple of of the form
        .. math:: \\alpha, r, p, \\tau, E

    Where

    - :math:`\\alpha` is the `repeating event`
    - :math:`r` is the number of repetitions of the event, called the `cycle length`
    - :math:`p` is the inter-occurrence distance, called the `cycle period`
    - :math:`\\tau` is the index of the first occurrence, called the `cycle starting point`
    - :math:`E` is a list of :math:`r - 1` signed integer offsets, i.e `cycle shift corrections`


    Parameters
    ----------
    max_length: int, default=20
        maximum length for a candidate cycle, when running the dynamic programming heuristic
    keep_residuals: bool, default=False
        Either to keep track of residuals (occurrences not covered by any cycle) or not.
        Residuals are required for a lossless reconstruction of the original data.
    n_jobs : int, default=1
        The number of jobs to use for the computation. Each single event is attributed a job
        to discover potential cycles.
        Threads are preffered over processes.

    Examples
    --------
    >>> from skmine.periodic import PeriodicCycleMiner
    >>> import pandas as pd
    >>> S = pd.Series("ring_a_bell", [10, 20, 32, 40, 60, 79, 100, 240])
    >>> pcm = PeriodicCycleMiner().fit(S)
    >>> pcm.discover()
                   start  length  period       cost
    ring_a_bell 1     10       3      11  23.552849
                0     40       4      20  24.665780

    References
    ----------
    .. [1]
        Galbrun, E & Cellier, P & Tatti, N & Termier, A & Crémilleux, B
        "Mining Periodic Pattern with a MDL Criterion"
    """

    def __init__(self, *, max_length=20, keep_residuals=False, n_jobs=1):
        self.miners_ = dict()
        self.is_datetime_ = None
        self.n_zeros_ = 0
        self.n_jobs = n_jobs
        self.max_length = max_length
        self.keep_residuals = keep_residuals

    def fit(self, S):
        """fit PeriodicCycleMiner on data logs

        This generate new candidate cycles and evaluate them.
        Residual occurrences are stored as an internal attribute,
        for later reconstruction (MDL is lossless)

        Parameters
        -------
        S: pd.Series
            logs, represented as a pandas Series
            This pandas Series must have an index of type in
            (pd.DatetimeIndex, pd.RangeIndex, pd.Int64Index)
        """
        if not isinstance(S, pd.Series):
            raise TypeError("S must be a pandas Series")

        if not isinstance(S.index, INDEX_TYPES):
            raise TypeError(f"S must have an index with a type amongst {INDEX_TYPES}")

        self.is_datetime_ = isinstance(S.index, pd.DatetimeIndex)

        if S.index.duplicated().any():
            warnings.warn("found duplicates in input sequence, removing them")
            S = S.groupby(S.index).first()

        S = S.copy()
        S.index, self.n_zeros_ = _remove_zeros(S.index.astype("int64"))
        # TODO : do this in SingleEventCycleMiner?

        n_occs_tot = S.shape[0]
        alpha_groups = S.groupby(S.values).groups
        miners = {
            k: SingleEventCycleMiner(self.max_length, self.keep_residuals, n_occs_tot)
            for k in alpha_groups.keys()
        }
        self.miners_ = {k: miners[k].fit(v) for k, v in alpha_groups.items()}

        for (event, miner,) in self.miners_.items():
            if "tids" in miner.cycles_.columns:
                # FIXME: this is highly inefficient
                miner.cycles_.tids = miner.cycles_.tids.map(
                    lambda tids: Bitmap(
                        np.searchsorted(S.index, alpha_groups[event][tids])
                    )
                )

        return self

    evaluate = SingleEventCycleMiner.evaluate

    def discover(self, shifts=False, tids=False):
        """Return cycles as a pandas DataFrame, with 3 columns,
        with a 2-level multi-index: the first level mapping events,
        and the second level being positional

        Returns
        -------
        pd.DataFrame
            DataFrame with the following columns
                ==========  ======================================
                start       when the cycle starts
                length      number of occurrences in the event
                period      inter-occurrence delay
                dE          shift corrections, if shifts=True
                tids        Transactions ids covered, if tids=True
                cost        MDL cost
                ==========  ======================================

        Example
        -------
        >>> from skmine.periodic import PeriodicCycleMiner
        >>> import pandas as pd
        >>> S = pd.Series("ring", [10, 20, 32, 40, 60, 79, 100, 240])
        >>> pcm = PeriodicCycleMiner().fit(S)
        >>> pcm.discover()
                start  length  period       cost
        ring 1     10       3      11  23.552849
             0     40       4      20  24.665780
        """
        all_cols = ["start", "length", "period", "cost"]
        if tids:
            all_cols.extend(["tids"])
        if shifts:
            all_cols.extend(["dE"])

        series = list()
        for miner in self.miners_.values():
            disc = miner.discover()
            if not disc.empty:
                series.append(disc)

        if not series:
            return pd.DataFrame()  # FIXME
        cycles = pd.concat(series, keys=self.miners_.keys())[all_cols]
        cycles.loc[:, ["start", "period"]] = cycles[["start", "period"]] * (
            10 ** self.n_zeros_
        )

        if shifts:
            cycles.loc[:, "dE"] = cycles.dE.map(np.array) * (10 ** self.n_zeros_)

        if self.is_datetime_:
            cycles.loc[:, "start"] = cycles.start.astype("datetime64[ns]")
            cycles.loc[:, "period"] = cycles.period.astype("timedelta64[ns]")

        return cycles

    def reconstruct(self):
        """Reconstruct the original occurrences from the current cycles.
        Residuals will also be included, as the compression scheme is lossless

        Denoting as :math:`\sigma(E)` the sum of the shift corrections for a cycle
        :math:`C`, we have

        .. math::
            \Delta(C)=(r-1) p+\sigma(E)

        Returns
        -------
        pd.Series
            The reconstructed dataset

        Notes
        -----
        The index of the resulting pd.Series will not be sorted
        """
        series = [
            pd.Series(event, index=miner.reconstruct())
            for event, miner in self.miners_.items()
        ]
        S = pd.concat(series)
        S.index *= 10 ** self.n_zeros_
        if self.is_datetime_:
            S.index = S.index.astype("datetime64[ns]")

        # TODO : this is due to duplicate event in the cycles, handle this in .evaluate()
        S = S.groupby(S.index).first()

        return S

    def generate_candidates(self, S):
        """
        Parameters
        ----------
        S: pd.Index or numpy.ndarray
            Series of occurrences for a specific event

        Returns
        -------
        dict[object, list[np.ndarray]]
            A dict, where each key is an event and each value a list of batch of candidates.
            Batches are sorted in inverse order of width,
            so that we consider larger candidate cycles first.
        """
        # TODO only for InteractiveMode

    def get_residuals(self):
        """Get the residual events, i.e events not covered by any cycle

        It is the complementary function to `discover`

        Returns
        -------
        pd.Series
            residual events
        """
        series = [
            pd.Series(event, index=miner.residuals_)
            for event, miner in self.miners_.items()
        ]
        residuals = pd.concat(series)
        if not residuals.empty:
            residuals.index *= 10 ** self.n_zeros_
        if self.is_datetime_:
            residuals.index = residuals.index.astype("datetime64[ns]")
        return residuals
